from functools import wraps


def fetch_property(loading_func: callable, should_load: callable = None):
    def _decorator(wrapped: callable):
        @wraps(wrapped)
        def wrapper(self, *args, **kwargs):
            if should_load is None or should_load(self):
                loading_func(self)
            return wrapped(self, *args, **kwargs)
        return wrapper
    return _decorator


def exception_handler(handler: callable) -> callable:
    def _decorator(wrapped: callable) -> callable:
        @wraps(wrapped)
        def wrapper(self, *args, **kwargs):
            try:
                return wrapped(self, *args, **kwargs)
            except Exception as exc:
                handler(self, exc)
        return wrapper
    return _decorator



from enum import Enum


class Tower(Enum):
    QUINCY = "Quincy"
    GWENDOLIN = "Gwendolin"
    STRIKER_JONES = "Striker Jones"
    OBYN = "Obyn Greenfoot"
    CHURCHILL = "Captain Churchill"
    BENJAMIN = "Benjamin"
    EZILI = "Ezili"
    PAT_FUSTY = "Pat Fusty"
    SAUDA = "Sauda"
    PSI = "Psi"
    GERALDO = "Geraldo"
    BRICKELL = "Admiral Brickell"
    ETIENNE = "Etienne"
    ADORA = "Adora"

    DART_MONKEY = "Dart Monkey"
    BOOMERANG_MONKEY = "Boomerang Monkey"
    TACK_SHOOTER = "Tack Shooter"
    BOMB_SHOOTER = "Bomb Shooter"
    GLUE_GUNNER = "Glue Gunner"
    ICE_MONKEY = "Ice Monkey"

    SNIPER_MONKEY = "Sniper Monkey"
    MONKEY_BUCCANEER = "Monkey Buccaneer"
    MONKEY_SUB = "Monkey Sub"
    DARTLING_GUNNER = "Dartling Gunner"
    MONKEY_ACE = "Monkey Ace"
    HELI_PILOT = "Heli Pilot"
    MORTAR_MONKEY = "Mortar Monkey"

    WIZARD_MONKEY = "Wizard Monkey"
    NINJA_MONKEY = "Ninja Monkey"
    SUPER_MONKEY = "Super Monkey"
    DRUID = "Druid"
    ALCHEMIST = "Alchemist"

    MONKEY_VILLAGE = "Monkey Village"
    BANANA_FARM = "Banana Farm"
    ENGINEER_MONKEY = "Engineer Monkey"
    SPIKE_FACTORY = "Spike Factory"

    @staticmethod
    def from_string(value: str) -> "Tower":
        tower_switch = {
            "Quincy": Tower.QUINCY,
            "Gwendolin": Tower.GWENDOLIN,
            "StrikerJones": Tower.STRIKER_JONES,
            "ObynGreenfoot": Tower.OBYN,
            "CaptainChurchill": Tower.CHURCHILL,
            "Benjamin": Tower.BENJAMIN,
            "Ezili": Tower.EZILI,
            "PatFusty": Tower.PAT_FUSTY,
            "Sauda": Tower.SAUDA,
            "Psi": Tower.PSI,
            "Geraldo": Tower.GERALDO,
            "AdmiralBrickell": Tower.BRICKELL,
            "Etienne": Tower.ETIENNE,
            "Adora": Tower.ADORA,
            "DartMonkey": Tower.DART_MONKEY,
            "BoomerangMonkey": Tower.BOOMERANG_MONKEY,
            "TackShooter": Tower.TACK_SHOOTER,
            "BombShooter": Tower.BOMB_SHOOTER,
            "GlueGunner": Tower.GLUE_GUNNER,
            "IceMonkey": Tower.ICE_MONKEY,
            "SniperMonkey": Tower.SNIPER_MONKEY,
            "MonkeyBuccaneer": Tower.MONKEY_BUCCANEER,
            "MonkeySub": Tower.MONKEY_SUB,
            "DartlingGunner": Tower.DARTLING_GUNNER,
            "MonkeyAce": Tower.MONKEY_ACE,
            "HeliPilot": Tower.HELI_PILOT,
            "MortarMonkey": Tower.MORTAR_MONKEY,
            "WizardMonkey": Tower.WIZARD_MONKEY,
            "NinjaMonkey": Tower.NINJA_MONKEY,
            "SuperMonkey": Tower.SUPER_MONKEY,
            "Druid": Tower.DRUID,
            "Alchemist": Tower.ALCHEMIST,
            "MonkeyVillage": Tower.MONKEY_VILLAGE,
            "BananaFarm": Tower.BANANA_FARM,
            "EngineerMonkey": Tower.ENGINEER_MONKEY,
            "SpikeFactory": Tower.SPIKE_FACTORY
        }
        return tower_switch[value] if value in tower_switch else None

    def is_hero(self):
        return self in [
            Tower.QUINCY, Tower.GWENDOLIN, Tower.STRIKER_JONES, Tower.OBYN, Tower.CHURCHILL, Tower.BENJAMIN,
            Tower.EZILI, Tower.PAT_FUSTY, Tower.SAUDA, Tower.PSI, Tower.GERALDO, Tower.BRICKELL, Tower.ETIENNE,
            Tower.ADORA
        ]



import unittest
from datetime import datetime
import bloonspy
from bloonspy import btd6


class TestChallenge(unittest.TestCase):
    def test_challenge(self) -> None:
        """
        Test that a challenge is loaded correctly.
        """
        challenge_code = "ZMDCDTB"
        challenge = btd6.Challenge(challenge_code)

        expected_results = [
            ("id", "ZMDCDTB"),
            ("name", "CT10 Tile CBA"),
            ("created_at", datetime.fromtimestamp(int(1671727285722/1000))),
            ("game_version", btd6.GameVersion(34, 3)),
            ("gamemode", btd6.Gamemode(btd6.Difficulty.MEDIUM, btd6.Mode.STANDARD)),
            ("disable_double_cash", True),
            ("disable_monkey_knowledge", False),
            ("disable_instas", False),
            ("disable_powers", True),
            ("disable_selling", False),
            ("disable_continues", True),
            ("starting_cash", 650),
            ("least_cash_used", bloonspy.Infinity()),
            ("least_tiers_used", 10),
            ("seed", 771803684),
            ("starting_lives", 150),
            ("max_lives", 9999),
            ("start_round", 1),
            ("end_round", 51),
            ("max_towers", 9999),
            ("max_paragons", 10),
            ("round_sets", ["default"]),
            # All values are defaults on this challenge anyway
            ("modifiers", btd6.ChallengeModifier()),
        ]
        for attr_name, attr_expected_value in expected_results:
            self.assertEqual(
                getattr(challenge, attr_name), attr_expected_value,
                msg=f"Asserting challenge.{attr_name}"
            )

        check_instance = [
            ("wins", int), ("plays", int), ("losses", int), ("upvotes", int), ("plays_unique", int),
            ("wins_unique", int), ("losses_unique", int)
        ]
        for attr_name, attr_type in check_instance:
            self.assertIsInstance(
                getattr(challenge, attr_name), attr_type,
                msg=f"Assert if challenge.{attr_name} is {attr_type}"
            )

        no_restrictions = btd6.TowerRestriction(
            max_towers=btd6.Infinity(),
            top_path_blocked=0,
            middle_path_blocked=0,
            bottom_path_blocked=0,
        )
        towers_allowed = {
            btd6.Tower.GWENDOLIN: btd6.Restriction(max_towers=1),
            btd6.Tower.DART_MONKEY: no_restrictions,
            btd6.Tower.GLUE_GUNNER: no_restrictions,
            btd6.Tower.SNIPER_MONKEY: no_restrictions,
            btd6.Tower.MONKEY_ACE: no_restrictions,
            btd6.Tower.HELI_PILOT: no_restrictions,
            btd6.Tower.MORTAR_MONKEY: no_restrictions,
            btd6.Tower.BANANA_FARM: no_restrictions,
        }
        for tower in challenge.towers.keys():
            self.assertIn(
                tower, towers_allowed.keys(),
                msg=f"Assert if {tower} is allowed"
            )
            self.assertEqual(
                towers_allowed[tower], challenge.towers[tower],
                msg=f"Assert if {towers_allowed[tower]} {tower} are available"
            )

        for power in btd6.Power:
            self.assertIn(
                power, challenge.powers.keys(),
                msg=f"Assert if {power.value} appears in the powers list"
            )
            self.assertEqual(
                challenge.powers[power], btd6.Infinity(),
                msg=f"Assert if {power.value} is allowed infinite times"
            )

    def test_challenge_fail(self) -> None:
        """
        Test that the correct exception is raised when attempting to fetch a nonexistant challenge.
        """
        challenge_code = "blatantlywrongchallengecode"
        correct_exception = False
        try:
            btd6.Challenge(challenge_code, eager=True)
        except btd6.NotFound:
            correct_exception = True
        self.assertTrue(correct_exception, msg="Wrong challenges IDs should raise bloonspy.exceptions.NotFound")


if __name__ == '__main__':
    unittest.main()



import unittest
from datetime import datetime
import requests
import bloonspy
from bloonspy import btd6


class TestCt(unittest.TestCase):
    def test_ct(self) -> None:
        """
        Test that a CT event event is loaded correctly.
        """
        ct_list = requests.get("https://data.ninjakiwi.com/btd6/ct")
        ct_id = ct_list.json()["body"][0]["id"]
        ct = btd6.ContestedTerritoryEvent(ct_id)

        check_instance = [
            ("name", str), ("start", datetime), ("end", datetime), ("id", str), ("total_scores_player", int),
            ("total_scores_team", int), ("event_number", int)
        ]
        for attr_name, attr_type in check_instance:
            self.assertIsInstance(getattr(ct, attr_name), attr_type,
                                  msg=f"Assert if ContestedTerritoryEvent.{attr_name} is {attr_type}")

    def test_ct_fail(self) -> None:
        """
        Test that the correct exception is raised when attempting to fetch a nonexistant/expired CT event.
        """
        ct_id = "l76rtr72"
        correct_exception = False
        try:
            btd6.ContestedTerritoryEvent(ct_id, eager=True)
        except btd6.NotFound:
            correct_exception = True
        self.assertTrue(correct_exception, msg="Wrong/expired CT IDs should raise bloonspy.exceptions.NotFound")


if __name__ == '__main__':
    unittest.main()



import unittest
from datetime import datetime
import bloonspy
from bloonspy import btd6


class TestTeam(unittest.TestCase):
    def test_team(self) -> None:
        """
        Test that a team is loaded correctly.
        """
        team_id = "9fbd42d9d995ada44c4188180975e57e990049b59917893b"
        team = btd6.Team(team_id)

        expected_results = [
            ("id", "9fbd42d9d995ada44c4188180975e57e990049b59917893b"),
            ("name", "PANDEMONIUM DEMONS"),
        ]
        for attr_name, attr_expected_value in expected_results:
            self.assertEqual(getattr(team, attr_name), attr_expected_value,
                             msg=f"Asserting Team.{attr_name}")

        self.assertGreater(team.member_count, 0,
                           msg=f"Check Team.member_count within lower bound")
        self.assertLessEqual(team.member_count, 15,
                             msg=f"Check Team.member_count within higher bound")

        check_instance = [
            ("status", btd6.TeamStatus), ("icon", btd6.Asset), ("banner", btd6.Asset),
            ("frame", btd6.Asset),
        ]
        for attr_name, attr_type in check_instance:
            self.assertIsInstance(getattr(team, attr_name), attr_type,
                                  msg=f"Assert if Team.{attr_name} is {attr_type}")

    def test_team_fail(self) -> None:
        """
        Test that the correct exception is raised when attempting to fetch a nonexistant team.
        """
        team_id = "blatantlywrongteamid"
        correct_exception = False
        try:
            btd6.Team(team_id, eager=True)
        except btd6.NotFound:
            correct_exception = True
        self.assertTrue(correct_exception, msg="Wrong team IDs should raise bloonspy.exceptions.NotFound")


if __name__ == '__main__':
    unittest.main()



import unittest
from datetime import datetime
import requests
import bloonspy
from bloonspy import btd6


class TestOdyssey(unittest.TestCase):
    def test_odyssey(self) -> None:
        """
        Test that an odyssey event is loaded correctly.
        """
        odysseys = requests.get("https://data.ninjakiwi.com/btd6/odyssey")
        odyssey_id = odysseys.json()["body"][0]["id"]
        odyssey = btd6.OdysseyEvent(odyssey_id)

        check_instance = [
            ("name", str), ("start", datetime), ("end", datetime), ("id", str)
        ]
        for attr_name, attr_type in check_instance:
            self.assertIsInstance(getattr(odyssey, attr_name), attr_type,
                                  msg=f"Assert if OdysseyEvent.{attr_name} is {attr_type}")

        easy_mode = odyssey.easy()
        medium_mode = odyssey.medium(eager=True)
        hard_mode = odyssey.hard()

        check_instance = [
            ("name", str), ("difficulty", btd6.OdysseyDifficulty), ("is_extreme", bool), ("max_monkey_seats", int),
            ("max_boat_seats", int), ("max_power_slots", int), ("starting_lives", int)
        ]
        for attr_name, attr_type in check_instance:
            self.assertIsInstance(getattr(hard_mode, attr_name), attr_type,
                                  msg=f"Assert if Odyssey.{attr_name} is {attr_type}")

        for power in medium_mode.available_powers.keys():
            self.assertIsInstance(power, btd6.Power,
                                  msg=f"Assert if Odyssey.available_powers is Dict[Power, int]")
            self.assertIsInstance(medium_mode.available_powers[power], int,
                                  msg=f"Assert if Odyssey.available_powers is Dict[Power, int]")

        for power in medium_mode.default_powers.keys():
            self.assertIsInstance(power, btd6.Power,
                                  msg=f"Assert if Odyssey.default_powers is Dict[Power, int]")
            self.assertIsInstance(medium_mode.available_powers[power], int,
                                  msg=f"Assert if Odyssey.default_powers is Dict[Power, int]")

        for tower in medium_mode.default_towers.keys():
            self.assertIsInstance(tower, btd6.Tower,
                                  msg=f"Assert if Odyssey.default_towers is Dict[Tower, int]")
            self.assertIsInstance(medium_mode.default_towers[tower], int,
                                  msg=f"Assert if Odyssey.default_towers is Dict[Tower, int]")

        for power in medium_mode.available_towers.keys():
            self.assertIsInstance(power, btd6.Tower,
                                  msg=f"Assert if Odyssey.available_towers is Dict[Tower, Restriction]")
            self.assertIsInstance(medium_mode.available_towers[power], btd6.Restriction,
                                  msg=f"Assert if Odyssey.available_towers is Dict[Tower, Restriction]")

        maps = easy_mode.maps()
        self.assertGreater(len(maps), 0)
        for island in maps:
            self.assertIsInstance(island, btd6.Challenge,
                                  msg="Assert if Odyssey.maps returns List[Challenge]")

    def test_odyssey_fail(self) -> None:
        """
        Test that the correct exception is raised when attempting to fetch a nonexistant/expired odyssey.
        """
        team_id = "blatantlywrongodysseyid"
        correct_exception = False
        try:
            btd6.OdysseyEvent(team_id, eager=True)
        except btd6.NotFound:
            correct_exception = True
        self.assertTrue(correct_exception, msg="Wrong odyssey IDs should raise bloonspy.exceptions.NotFound")


if __name__ == '__main__':
    unittest.main()



import unittest
import random
from bloonspy import btd6, Client


class TestCtLeaderboard(unittest.TestCase):
    def test_ct_leaderboard(self) -> None:
        """
        Test a CT event's leaderboard.
        """
        cts = Client.contested_territories()
        ct_event = cts[1]
        # Get #51-#100
        ct_leaderboard_team = ct_event.leaderboard_team(pages=2, start_from_page=2)
        ct_leaderboard_player = ct_event.leaderboard_player(pages=2, start_from_page=2)

        self.assertGreater(len(ct_leaderboard_player), 0)
        some_player = ct_leaderboard_player[random.randint(0, len(ct_leaderboard_player)-1)]
        self.assertIsInstance(some_player, btd6.CtPlayer,
                              msg=f"Assert if result is CtPlayer")

        check_instance = [
            ("name", str), ("score", int), ("achievements", int), ("boss_normal_medals", btd6.EventMedals)
        ]
        for attr_name, attr_type in check_instance:
            self.assertIsInstance(getattr(some_player, attr_name), attr_type,
                                  msg=f"Assert if CtPlayer.{attr_name} is {attr_type}")

        self.assertGreater(len(ct_leaderboard_team), 0)
        some_team = ct_leaderboard_team[random.randint(0, len(ct_leaderboard_team)-1)]
        self.assertIsInstance(some_team, btd6.CtTeam,
                              msg=f"Assert if result is CtTeam")

        check_instance = [
            ("name", str), ("score", int), ("member_count", int), ("status", btd6.TeamStatus)
        ]
        for attr_name, attr_type in check_instance:
            self.assertIsInstance(getattr(some_team, attr_name), attr_type,
                                  msg=f"Assert if CtTeam.{attr_name} is {attr_type}")


if __name__ == '__main__':
    unittest.main()



from typing import Dict, Any, List, Tuple


def rename_keys(dictionary: Dict[str, Any], keys: List[Tuple[str, str]]) -> Dict[str, Any]:
    new_dict = {}
    for old_key, new_key in keys:
        item = dictionary
        current_keys = old_key.split(".")
        while len(current_keys) > 0:
            key = current_keys.pop(0)
            if key not in item:
                item = None
                break
            item = item[key]
        if item is not None:
            new_dict[new_key] = item

    return new_dict


def has_all_keys(dictionary: Dict[str, Any], key_list: List[str]):
    for key in key_list:
        if key not in dictionary.keys():
            return False
    return True



import unittest
from bloonspy import btd6, Client


class TestFetchOdyssey(unittest.TestCase):
    def test_odyssey(self) -> None:
        """
        Test getting a random odyssey.
        """
        odysseys = Client.odysseys()
        some_odyssey = odysseys[0]
        self.assertIsInstance(some_odyssey, btd6.OdysseyEvent,
                              msg=f"Assert if Client.odysseys returns List[OdysseyEvent]")

        some_other_odyssey = Client.get_odyssey(odysseys[1].id)
        self.assertIsInstance(some_other_odyssey, btd6.OdysseyEvent,
                              msg=f"Assert if Client.get_odyssey returns OdysseyEvent")


if __name__ == '__main__':
    unittest.main()



from dataclasses import dataclass, field


@dataclass(kw_only=True)
class EventMedals:
    """Medal set for events."""
    first: int = field(default=0)  #: Number of first place finishes.
    second: int = field(default=0)  #: Number of second place finishes.
    third: int = field(default=0)  #: Number of third place finishes.
    top_50: int = field(default=0)  #: Number of Top 50 finishes.
    top_1_percent: int = field(default=0)  #: Number of Top 1% finishes.
    top_10_percent: int = field(default=0)  #: Number of Top 10% finishes.
    top_25_percent: int = field(default=0)  #: Number of Top 25% finishes.
    top_50_percent: int = field(default=0)  #: Number of Top 50% finishes.
    top_75_percent: int = field(default=0)  #: Number of Top 75% finishes.


@dataclass(kw_only=True)
class CTLocalMedals:
    """Medal set for Contested Territory (local leaderboard)."""
    first: int = field(default=0)  #: Number of first place finishes.
    second: int = field(default=0)  #: Number of second place finishes.
    third: int = field(default=0)  #: Number of third place finishes.
    top_10: int = field(default=0)  #: Number of Top 10 finishes.
    top_20: int = field(default=0)  #: Number of Top 20 finishes.
    top_40: int = field(default=0)  #: Number of Top 40 finishes.
    top_60: int = field(default=0)  #: Number of Top 60 finishes.


@dataclass(kw_only=True)
class CTGlobalMedals:
    """Medal set for Contested Territory (global leaderboard)."""
    top_25: int = field(default=0)  #: Number of Top 25 finishes.
    top_100: int = field(default=0)  #: Number of Top 100 finishes.
    top_1_percent: int = field(default=0)  #: Number of Top 1% finishes.
    top_10_percent: int = field(default=0)  #: Number of Top 10% finishes.
    top_25_percent: int = field(default=0)  #: Number of Top 25% finishes.
    top_50_percent: int = field(default=0)  #: Number of Top 50% finishes.
    top_75_percent: int = field(default=0)  #: Number of Top 75% finishes.


@dataclass(kw_only=True)
class MapMedals:
    """Medal set for maps."""
    easy: int = field(default=0)  #: Number of medals for Easy - Standard.
    primary_only: int = field(default=0)  #: Number of medals for Easy - Primary only.
    deflation: int = field(default=0)  #: Number of medals for Easy - Deflation.
    medium: int = field(default=0)  #: Number of medals for Medium - Standard.
    military_only: int = field(default=0)  #: Number of medals for Medium - Military only.
    apopalypse: int = field(default=0)  #: Number of medals for Medium - Apopalypse.
    reverse: int = field(default=0)  #: Number of medals for Medium - Reverse.
    hard: int = field(default=0)  #: Number of medals for Hard - Standard.
    magic_only: int = field(default=0)  #: Number of medals for Hard - Magic only.
    double_hp_moabs: int = field(default=0)  #: Number of medals for Hard - Double HP MOABs.
    half_cash: int = field(default=0)  #: Number of medals for Hard - Half Cash.
    alternate_bloons_rounds: int = field(default=0)  #: Number of medals for Hard - Alternate Bloons Rounds.
    impoppable: int = field(default=0)  #: Number of medals for Hard - Impoppable.
    chimps_red: int = field(default=0)  #: Number of medals for Hard - CHIMPS (red).
    chimps_black: int = field(default=0)  #: Number of medals for Hard - CHIMPS (black).



from enum import Enum


class Power(Enum):
    SMS = "Super Monkey Storm"
    MONKEY_BOOST = "Monkeyboost"
    THRIVE = "Thrive"
    TIME_STOP = "Time Stop"
    CASH_DROP = "Cash Drop"
    PONTOON = "Pontoon"
    ROAD_SPIKES = "Road Spikes"
    GLUE_TRAP = "Glue Trap"
    MOAB_MINE = "Moab Mine"
    CAMO_TRAP = "Camo Trap"
    PORTABLE_LAKE = "Portable Lake"
    TECH_BOT = "Tech Bot"
    ENERGISING_TOTEM = "Energising Totem"
    BANANA_FARMER = "Banana Farmer"

    @staticmethod
    def from_string(power: str) -> "Power":
        power_switch = {
            "SuperMonkeyStorm": Power.SMS,
            "MonkeyBoost": Power.MONKEY_BOOST,
            "Thrive": Power.THRIVE,
            "DartTime": Power.TIME_STOP,
            "CashDrop": Power.PONTOON,
            "BananaFarmer": Power.BANANA_FARMER,
            "Pontoon": Power.CASH_DROP,
            "RoadSpikes": Power.ROAD_SPIKES,
            "GlueTrap": Power.GLUE_TRAP,
            "MoabMine": Power.MOAB_MINE,
            "CamoTrap": Power.CAMO_TRAP,
            "PortableLake": Power.PORTABLE_LAKE,
            "TechBot": Power.TECH_BOT,
            "EnergisingTotem": Power.ENERGISING_TOTEM,
        }
        power = power_switch[power] if power in power_switch else None
        return power



from datetime import datetime, timedelta
from concurrent.futures import ThreadPoolExecutor
from typing import List, Dict, Any
from ...utils.decorators import fetch_property, exception_handler
from ...utils.dictionaries import has_all_keys
from ...utils.api import get, get_lb_page
from ...exceptions import NotFound
from .Challenge import Challenge
from .User import User


class RacePlayer(User):
    """An user who played a race and is now on the leaderboard.
    Inherits from :class:`~bloonspy.model.btd6.User`.
    """
    def __init__(self, user_id: str, name: str, score: int, submission_time: int, **kwargs):
        super().__init__(user_id, **kwargs)
        self._name = name
        self._score = timedelta(seconds=int(score/1000))
        self._submission_time = datetime.fromtimestamp(int(submission_time/1000))

    @property
    def name(self) -> str:
        """The name of the user."""
        return self._name

    @property
    def score(self) -> timedelta:
        """The time the user got."""
        return self._score

    @property
    def submission_time(self) -> datetime:
        """The time the user's score was submitted at."""
        return self._submission_time


class Race(Challenge):
    """A race event. Inherits from :class:`~bloonspy.model.btd6.Challenge`."""

    endpoint = "/btd6/races/{}/metadata"
    event_endpoint = "/btd6/races"
    lb_endpoint = "/btd6/races/{}/leaderboard"

    def __init__(self, race_id: str, eager: bool = True, race_json: Dict[str, Any] = None):
        super().__init__(race_id, eager=eager)
        self._start = datetime.fromtimestamp(0)
        self._end = datetime.fromtimestamp(0)
        self._total_scores = 0
        self._race_loaded = eager
        if race_json and has_all_keys(race_json, ["name", "start", "end", "totalScores"]):
            self._parse_race(race_json)
        if eager and not self._race_loaded:
            self._load_race()

    def _handle_exceptions(self, exception: Exception) -> None:
        error_msg = str(exception)
        if error_msg == "No Race with that ID exists":
            raise NotFound(error_msg)

    @exception_handler(Challenge.handle_exceptions)
    def _load_race(self, only_if_unloaded: bool = True) -> None:
        if self._race_loaded and only_if_unloaded:
            return

        self._race_loaded = False

        race_list = get(self.event_endpoint)
        for race in race_list:
            if race["id"] == self._id:
                self._parse_race(race)
                return

        raise NotFound("No Race with that ID exists")

    def _parse_race(self, data: Dict[str, Any]) -> None:
        self._data["name"] = data["name"]
        self._start = datetime.fromtimestamp(data["start"]/1000)
        self._end = datetime.fromtimestamp(data["end"]/1000)
        self._total_scores = data["totalScores"]
        self._race_loaded = True

    @property
    @fetch_property(_load_race)
    def name(self) -> str:
        """The name of the race."""
        return self._data["name"]

    @property
    @fetch_property(_load_race)
    def start(self) -> datetime:
        """When the event starts."""
        return self._start

    @property
    @fetch_property(_load_race)
    def end(self) -> datetime:
        """When the event ends."""
        return self._end

    @property
    @fetch_property(_load_race)
    def total_scores(self) -> int:
        """Number of users who played and submitted a score."""
        return self._total_scores

    @exception_handler(Challenge.handle_exceptions)
    def leaderboard(self, pages: int = 1, start_from_page: int = 0) -> List[RacePlayer]:
        """Get a page of the leaderboard for this event.

        .. note::
           The returned :class:`~bloonspy.model.btd6.RacePlayer` objects will only
           have the properties :attr:`~bloonspy.model.Loadable.id`, :attr:`~bloonspy.model.btd6.Player.name`,
           :attr:`~bloonspy.model.RacePlayer.score`, and :attr:`~bloonspy.model.RacePlayer.submission_time` loaded.

        :param pages: Number of pages to fetch.
        :type pages: int
        :param start_from_page: The first page to fetch.
        :type start_from_page: int

        :return: A list of players in the leaderboard.
        :rtype: List[:class:`~bloonspy.model.btd6.RacePlayer`]

        :raise ~bloonspy.exceptions.NotFound: If the race doesn't exist or is expired.
        """
        futures = []
        with ThreadPoolExecutor(max_workers=10) as executor:
            for page_num in range(start_from_page, start_from_page + pages):
                futures.append(executor.submit(get_lb_page, self.lb_endpoint.format(self._id), page_num))

        race_players = []
        for page in futures:
            for player in page.result():
                race_players.append(RacePlayer(
                    player["profile"].split("/")[-1], player["displayName"], player["score"], player["submissionTime"]
                ))

        return race_players





class BloonsException(Exception):
    """Superclass for all Bloons exceptions"""
    pass


class NotFound(BloonsException):
    """The requested resource or event doesn't exist or has expired."""
    pass




from dataclasses import dataclass
from enum import Enum


class Difficulty(Enum):
    EASY = "Easy"
    MEDIUM = "Medium"
    HARD = "Hard"
    IMPOPPABLE = "Impoppable"


class Mode(Enum):
    STANDARD = "Standard"

    PRIMARY_ONLY = "Primary Only"
    DEFLATION = "Deflation"

    MILITARY_ONLY = "Military Only"
    REVERSE = "Reverse"
    APOPALYPSE = "Apopalypse"

    MAGIC_ONLY = "Magic Only"
    DOUBLE_HP_MOABS = "Double HP MOABs"
    HALF_CASH = "Half Cash"
    ALTERNATE_BLOONS_ROUNDS = "Alternate Bloons Rounds"

    CHIMPS = "CHIMPs"


@dataclass
class Gamemode:
    """Difficulty and mode for a game."""
    difficulty: Difficulty  #: The game's difficulty.
    mode: Mode  #: The game's mode.

    @staticmethod
    def from_strings(difficulty, mode) -> "Gamemode":
        difficulty_switch = {
            "Easy": Difficulty.EASY,
            "Medium": Difficulty.MEDIUM,
            "Hard": Difficulty.HARD,
            "Impoppable": Difficulty.IMPOPPABLE,
        }
        # TODO get the actual strings
        mode_switch = {
            "Standard": Mode.STANDARD,
            "PrimaryOnly": Mode.PRIMARY_ONLY,
            "Deflation": Mode.DEFLATION,
            "MilitaryOnly": Mode.MILITARY_ONLY,
            "Reverse": Mode.REVERSE,
            "Apopalypse": Mode.APOPALYPSE,
            "MagicOnly": Mode.MAGIC_ONLY,
            "DoubleHPMoabs": Mode.DOUBLE_HP_MOABS,
            "ABR": Mode.ALTERNATE_BLOONS_ROUNDS,
            "Chimps": Mode.CHIMPS,
        }

        found_difficulty = difficulty_switch[difficulty] if difficulty in difficulty_switch else None
        found_mode = mode_switch[mode] if mode in mode_switch else None
        return Gamemode(found_difficulty, found_mode)



import unittest
from datetime import datetime, timedelta
import requests
import random
from bloonspy import btd6, Client


class TestBossLeaderboard(unittest.TestCase):
    def test_boss_leaderboard(self) -> None:
        """
        Test fetching a boss' leaderboard.
        """
        bosses = Client.bosses()
        boss = bosses[1]
        # Get #51-#100
        boss_leaderboard = boss.standard().leaderboard(pages=2, start_from_page=2)

        self.assertGreater(len(boss_leaderboard), 0)
        some_player = boss_leaderboard[random.randint(0, len(boss_leaderboard)-1)]
        self.assertIsInstance(some_player, btd6.BossPlayer,
                              msg=f"Assert if result is BossPlayer")

        check_instance = [
            ("name", str), ("score", timedelta), ("submission_time", datetime),
            ("achievements", int), ("boss_normal_medals", btd6.EventMedals)
        ]
        for attr_name, attr_type in check_instance:
            self.assertIsInstance(getattr(some_player, attr_name), attr_type,
                                  msg=f"Assert if BossPlayer.{attr_name} is {attr_type}")

    # def test_boss_leaderboard_coop(self) -> None:
    #     """
    #     Test fetching a boss' coop leaderboard.
    #     """


if __name__ == '__main__':
    unittest.main()



import concurrent.futures
from typing import List
from concurrent.futures import ThreadPoolExecutor
from .utils.api import get
from .model.btd6 import \
    OdysseyEvent, \
    BossEvent, \
    Race, \
    ContestedTerritoryEvent, Team, \
    Challenge, ChallengeFilter, \
    User


class Client:
    """Client for all API calls.

    :param open_access_key: Your OAK for the Ninja Kiwi Open Data API.
    :type open_access_key: str
    """
    def __init__(self, open_access_key: str):
        self.__oak = open_access_key

    @staticmethod
    def odysseys() -> List[OdysseyEvent]:
        """Get a list of Odyssey events."""
        odysseys_data = get("/btd6/odyssey")
        odyssey_list = []
        for odyssey in odysseys_data:
            odyssey_list.append(OdysseyEvent(odyssey["id"], event_json=odyssey))
        return odyssey_list

    @staticmethod
    def get_odyssey(odyssey_id: str, eager: bool = True) -> OdysseyEvent:
        """Fetch a specific Odyssey by its ID.

        .. note::
           If lazy loaded, the returned :class:`~bloonspy.model.btd6.OdysseyEvent` object will only
           have the property :attr:`~bloonspy.model.Event.id` loaded.

        :param odyssey_id: The ID of the odyssey.
        :type odyssey_id: str
        :param eager: If `True`, it loads all of the data right away. Set it to `False`
            if you want to limit API calls and don't need all the data. For more information,
            please read `Lazy and Eager Loading <async.html#lazy-and-eager-loading>`_.
        :type eager: bool

        :raise ~bloonspy.exceptions.NotFound: If no odyssey with that ID is found.

        :return: The found odyssey.
        :rtype: ~bloonspy.model.btd6.OdysseyEvent
        """
        return OdysseyEvent(odyssey_id, eager=eager)

    @staticmethod
    def contested_territories() -> List[ContestedTerritoryEvent]:
        """Get a list of Contested Territory events."""
        ct_data = get("/btd6/ct")
        ct_list = []
        for ct in ct_data:
            ct_list.append(ContestedTerritoryEvent(ct["id"], event_json=ct))
        return ct_list

    @staticmethod
    def get_contested_territory(ct_id: str, eager: bool = True) -> ContestedTerritoryEvent:
        """Fetch a specific Contested Territory event by its ID.

        .. note::
           If lazy loaded, the returned :class:`~bloonspy.model.btd6.ContestedTerritoryEvent` object will only
           have the property :attr:`~bloonspy.model.Lodable.id` loaded.

        :param ct_id: The ID of the event.
        :type ct_id: str
        :param eager: If `True`, it loads all of the data right away. Set it to `False`
            if you want to limit API calls and don't need all the data. For more information,
            please read `Lazy and Eager Loading <async.html#lazy-and-eager-loading>`_.
        :type eager: bool

        :raise ~bloonspy.exceptions.NotFound: If no CT with that ID is found.

        :return: The found Contested Territory event.
        :rtype: ~bloonspy.model.btd6.ContestedTerritoryEvent
        """
        return ContestedTerritoryEvent(ct_id, eager=eager)

    @staticmethod
    def get_team(team_id: str) -> Team:
        """Fetch a specific team by its ID.

        :param team_id: The ID of the team.
        :type team_id: str

        :raise ~bloonspy.exceptions.NotFound: If no team with that ID is found.

        :return: The found team.
        :rtype: ~bloonspy.model.btd6.Team
        """
        return Team(team_id, eager=True)

    @staticmethod
    def races() -> List[Race]:
        """Get a list of Race events.

        .. note::
           The returned :class:`~bloonspy.model.btd6.Race` objects will only
           have the properties :attr:`~bloonspy.model.Loadable.id`, :attr:`~bloonspy.model.btd6.Race.name`,
           :attr:`~bloonspy.model.Race.start`, :attr:`~bloonspy.model.Race.end`, and
           :attr:`~bloonspy.model.Race.total_scores` loaded.
        """
        races_data = get("/btd6/races")
        race_list = []
        for race in races_data:
            race_list.append(Race(race["id"], race_json=race))
        return race_list

    @staticmethod
    def get_race(race_id: str, eager: bool = True) -> Race:
        """Fetch a specific Race by its ID.

        .. note::
           If lazy loaded, the returned :class:`~bloonspy.model.btd6.Race` objects will only
           have the properties :attr:`~bloonspy.model.Loadable.id` loaded.

        :param race_id: The ID of the race.
        :type race_id: str
        :param eager: If `True`, it loads all of the data right away. Set it to `False`
            if you want to limit API calls and don't need all the data. For more information,
            please read `Lazy and Eager Loading <async.html#lazy-and-eager-loading>`_.
        :type eager: bool

        :raise ~bloonspy.exceptions.NotFound: If no race with that ID is found.

        :return: The found race.
        :rtype: ~bloonspy.model.btd6.Race
        """
        return Race(race_id, eager=eager)

    @staticmethod
    def bosses() -> List[BossEvent]:
        """Get a list of Boss events."""
        bosses_data = get("/btd6/bosses")
        boss_list = []
        for boss in bosses_data:
            boss_list.append(BossEvent(boss["id"], event_json=boss))
        return boss_list

    @staticmethod
    def get_boss(boss_id: str, eager: bool = True) -> BossEvent:
        """Fetch a specific Boss event by its ID.

        .. note::
           If lazy loaded, the returned :class:`~bloonspy.model.btd6.BossEvent` objects will only
           have the property :attr:`~bloonspy.model.Loadable.id` loaded.

        :param boss_id: The boss ID.
        :type boss_id: str
        :param eager: If `True`, it loads all of the data right away. Set it to `False`
            if you want to limit API calls and don't need all the data. For more information,
            please read `Lazy and Eager Loading <async.html#lazy-and-eager-loading>`_.
        :type eager: bool

        :raise ~bloonspy.exceptions.NotFound: If no boss event with that ID is found.

        :return: The found Boss event.
        :rtype: ~bloonspy.model.btd6.BossEvent
        """
        return BossEvent(boss_id, eager=eager)

    @staticmethod
    def challenges(challenge_filter: ChallengeFilter, pages: int = 1, start_from_page: int = 1) -> List[Challenge]:
        """Get a list of challenges given a specific filter.
        
        .. note::
           The returned :class:`~bloonspy.model.btd6.Challenge` objects will only
           have the properties :attr:`~bloonspy.model.Loadable.id`, :attr:`~bloonspy.model.btd6.Challenge.name`,
           :attr:`~bloonspy.model.Challenge.created_at`, and :attr:`~bloonspy.model.Challenge.creator_id` loaded.

        :param challenge_filter: Which type of challenges you'd like to see.
        :type challenge_filter: ~bloonspy.model.btd6.ChallengeFilter
        :param pages: The number of pages to fetch.
        :type pages: int
        :param start_from_page: The page to start fetching from.
        :type start_from_page: int

        :return: A list of challenges (lazy loaded).
        :rtype: List[:class:`bloonspy.model.btd6.Challenge`]
        """

        challenge_list = []
        with ThreadPoolExecutor(max_workers=5) as executor:
            challenge_pages = []
            for page_num in range(start_from_page, start_from_page+pages):
                challenge_pages.append(executor.submit(
                    get, f"/btd6/challenges/filter/{challenge_filter.value}", {"page": page_num}
                ))
            for page in challenge_pages:
                for chlg in page.result():
                    challenge_list.append(Challenge(chlg["id"], name=chlg["name"], created_at=chlg["createdAt"],
                                                    creator_id=chlg["creator"].split("/")[-1]))
        # if eager:
        #     with ThreadPoolExecutor(max_workers=10) as executor:
        #         futures = []
        #         for challenge in challenge_list:
        #             futures.append(executor.submit(challenge.load_resource))
        #         concurrent.futures.wait(futures)

        return challenge_list

    @staticmethod
    def get_challenge(challenge_id: str) -> Challenge:
        """Fetch a specific challenge by its ID.

        :param challenge_id: The challenge ID.
        :type challenge_id: str

        :raise ~bloonspy.exceptions.NotFound: If no challenge with the given ID is found.

        :return: The found challenge.
        :rtype: ~bloonspy.model.btd6.Challenge
        """
        return Challenge(challenge_id, eager=True)

    @staticmethod
    def get_user(identifier: str) -> User:
        """Fetch a specific user by an identifier.

        :param identifier: The user ID, or its OAK.
        :type identifier: str

        :raise ~bloonspy.exceptions.NotFound: If no user with the given ID/OAK is found.

        :return: The found user.
        :rtype: ~bloonspy.model.btd6.User
        """
        return User(identifier, eager=True)



from .Boss import *
from .Challenge import *
from .ContestedTerritory import *
from .Gamemode import *
from .Medals import *
from .Odyssey import *
from .Power import *
from .Race import *
from .Restriction import *
from .Team import *
from .Tower import *
from .User import *



import unittest
from datetime import datetime, timedelta
import requests
import random
from bloonspy import btd6, Client


class TestRaceLeaderboard(unittest.TestCase):
    def test_race_leaderboard(self) -> None:
        """
        Test getting a race event's leaderboard.
        """
        races = Client.races()
        race = races[1]
        # Get #51-#100
        race_leaderboard = race.leaderboard(pages=2, start_from_page=2)

        self.assertGreater(len(race_leaderboard), 0)
        some_player = race_leaderboard[random.randint(0, len(race_leaderboard)-1)]
        self.assertIsInstance(some_player, btd6.RacePlayer,
                              msg=f"Assert if result is RacePlayer")

        check_instance = [
            ("name", str), ("score", timedelta), ("submission_time", datetime),
            ("achievements", int), ("boss_normal_medals", btd6.EventMedals)
        ]
        for attr_name, attr_type in check_instance:
            self.assertIsInstance(getattr(some_player, attr_name), attr_type,
                                  msg=f"Assert if RacePlayer.{attr_name} is {attr_type}")


if __name__ == '__main__':
    unittest.main()



from enum import Enum
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timedelta
from typing import List, Dict, Any
from ...utils.decorators import fetch_property, exception_handler
from ...utils.api import get, get_lb_page
from ..Loadable import Loadable
from ..Event import Event
from .Challenge import Challenge
from .User import User


class BossBloon(Enum):
    BLOONARIUS = "bloonarius"
    LYCH = "lych"
    VORTEX = "vortex"
    DREADBLOON = "dreadbloon"
    # BLASTAPOPOULOS = "blastapopoulos"

    @staticmethod
    def from_string(boss: str) -> "BossBloon":
        boss_switch = {
            "vortex": BossBloon.VORTEX,
            "bloonarius": BossBloon.BLOONARIUS,
            "lych": BossBloon.LYCH,
            "dreadbloon": BossBloon.DREADBLOON,
        }
        boss = boss_switch[boss] if boss in boss_switch else None
        return boss


class BossPlayer(User):
    """An user who played the boss event and submitted a ranked score.
    Inherits from :class:`~bloonspy.model.btd6.User`.
    """
    def __init__(self, user_id: str, name: str, score: int, submission_time: int, **kwargs):
        super().__init__(user_id, **kwargs)
        self._name = name
        self._score = timedelta(seconds=int(score/1000))
        self._submission_time = datetime.fromtimestamp(int(submission_time/1000))

    @property
    def name(self) -> str:
        """The name of the user."""
        return self._name

    @property
    def score(self) -> timedelta:
        """The time the user got."""
        return self._score

    @property
    def submission_time(self) -> datetime:
        """The time the user’s score was submitted at."""
        return self._submission_time


class Boss(Challenge):
    """A Boss challenge. Inherits from :class:`~bloonspy.model.btd6.Challenge`."""
    endpoint = "/btd6/bosses/{}/metadata/:difficulty:"
    lb_endpoint = "/btd6/bosses/{}/leaderboard/:difficulty:/{}"

    def __init__(self, boss_id: str, name: str, boss_bloon: BossBloon, total_scores: int, elite: bool,
                 eager: bool = True):
        self._is_elite = elite
        self.endpoint = self.endpoint.replace(":difficulty:", "elite" if self._is_elite else "standard")
        self.lb_endpoint = self.lb_endpoint.replace(":difficulty:", "elite" if self._is_elite else "standard")
        super().__init__(boss_id, eager=eager)
        self._data["name"] = name
        self._boss_bloon = boss_bloon
        self._total_scores = total_scores

    @property
    def boss_bloon(self) -> BossBloon:
        """The boss in this challenge."""
        return self._boss_bloon

    @property
    def total_scores(self) -> int:
        """The total scores submitted."""
        return self._total_scores

    @property
    def is_elite(self) -> bool:
        """`True` if the boss is Elite."""
        return self._is_elite

    def _get_lb_page(self, page_num: int, team_size: int):
        try:
            return get(self.lb_endpoint.format(self._id, team_size), params={"page": page_num})
        except Exception as exc:
            if str(exc) == "No Scores Available":
                return []
            raise exc

    @exception_handler(Loadable.handle_exceptions)
    def leaderboard(self, pages: int = 1, start_from_page: int = 0, team_size: int = 1) -> List[BossPlayer]:
        """Get a page of the leaderboard for this boss.

        .. note::
           The returned :class:`~bloonspy.model.btd6.BossPlayer` objects will only
           have the properties :attr:`~bloonspy.model.Loadable.id`, :attr:`~bloonspy.model.btd6.BossPlayer.name`,
           :attr:`~bloonspy.model.BossPlayer.score`, and :attr:`~bloonspy.model.BossPlayer.submission_time` loaded.

        :param pages: Number of pages to fetch.
        :type pages: int
        :param start_from_page: The first page to fetch.
        :type start_from_page: int
        :param team_size: The team size to get the leaderboard for.
        :type team_size: int

        :return: A list of players in the leaderboard.
        :rtype: List[:class:`~bloonspy.model.btd6.RacePlayer`]

        :raise ~bloonspy.exceptions.NotFound: If the boss doesn't exist or is expired.
        :raise ValueError: If `team_size` is less than 1 or more than 4.
        """
        if team_size not in range(1, 5):
            raise ValueError("team_size must be between 1 and 4")

        futures = []
        with ThreadPoolExecutor(max_workers=10) as executor:
            for page_num in range(start_from_page, start_from_page + pages):
                futures.append(executor.submit(get_lb_page, self.lb_endpoint.format(self._id, team_size), page_num))

        boss_players = []
        for page in futures:
            for player in page.result():
                boss_players.append(BossPlayer(
                    player["profile"].split("/")[-1], player["displayName"], player["score"], player["submissionTime"]
                ))

        return boss_players


class BossEvent(Event):
    """A boss event. Inherits from :class:`~bloonspy.model.Event`."""
    event_endpoint = "/btd6/bosses"
    event_dict_keys = ["name", "bossType", "bossTypeURL", "start", "end", "totalScores_standard",
                       "totalScores_elite"]
    event_name: str = "Boss"

    def _parse_event(self, data: Dict[str, Any]) -> None:
        self._data["boss_bloon"] = BossBloon.from_string(data["bossType"])
        self._data["boss_banner"] = data["bossTypeURL"]
        self._data["total_scores_standard"] = data["totalScores_standard"]
        self._data["total_scores_elite"] = data["totalScores_elite"]
        super()._parse_event(data)

    @property
    @fetch_property(Event.load_event, should_load=Event._should_load_property)
    def boss_bloon(self) -> BossBloon:
        """The boss bloon in this event."""
        return self._data["boss_bloon"]

    @property
    @fetch_property(Event.load_event, should_load=Event._should_load_property)
    def boss_banner(self) -> str:
        """The URL to the banner used to advertise the event."""
        return self._data["boss_banner"]

    @property
    @fetch_property(Event.load_event, should_load=Event._should_load_property)
    def total_scores_standard(self) -> int:
        """Total scores submitted in the standard boss."""
        return self._data["total_scores_standard"]

    @property
    @fetch_property(Event.load_event, should_load=Event._should_load_property)
    def total_scores_elite(self) -> int:
        """Total scores submitted in the elite boss."""
        return self._data["total_scores_elite"]

    def standard(self, eager: bool = False) -> Boss:
        """Get the standard boss challenge.

        .. note::
           If lazy loaded, the returned :class:`~bloonspy.model.btd6.Boss` object will only
           have the properties :attr:`~bloonspy.model.Loadable.id`, :attr:`~bloonspy.model.btd6.Challenge.name`,
           :attr:`~bloonspy.model.Boss.boss_bloon`, :attr:`~bloonspy.model.Boss.is_elite`,
           and :attr:`~bloonspy.model.Boss.total_scores` loaded.

        :param eager: If `True`, it loads all of the data right away. Set it to `False`
            if you want to limit API calls and don't need all the data. For more information,
            please read `Lazy and Eager Loading <async.html#lazy-and-eager-loading>`_.
        :type eager: bool
        :return: The standard boss event.
        :rtype: ~bloonspy.model.btd6.Boss"""
        return Boss(self.id, self.name, self.boss_bloon,
                    self.total_scores_standard, False, eager=eager)

    def elite(self, eager: bool = False) -> Boss:
        """Get the elite boss challenge.

        .. note::
           If lazy loaded, the returned :class:`~bloonspy.model.btd6.Boss` object will only
           have the properties :attr:`~bloonspy.model.Loadable.id`, :attr:`~bloonspy.model.btd6.Challenge.name`,
           :attr:`~bloonspy.model.Boss.boss_bloon`, :attr:`~bloonspy.model.Boss.is_elite`,
           and :attr:`~bloonspy.model.Boss.total_scores` loaded.

        :param eager: If `True`, it loads all of the data right away. Set it to `False`
            if you want to limit API calls and don't need all the data. For more information,
            please read `Lazy and Eager Loading <async.html#lazy-and-eager-loading>`_.
        :type eager: bool
        :return: The elite boss event.
        :rtype: ~bloonspy.model.btd6.Boss"""
        return Boss(self.id, self.name, self.boss_bloon,
                    self.total_scores_elite, True, eager=eager)



import unittest
from bloonspy import btd6, Client


class TestTeamOwner(unittest.TestCase):
    def test_team(self) -> None:
        """
        Test getting a team's owner.
        """
        team_id = "9fbd42d98ac2faa41d40884a5b21b577cb064ebacf168e3f"
        team = Client.get_team(team_id)
        owner = team.owner()
        self.assertIsNotNone(owner)
        self.assertIn(owner.name.lower(), ["thargos", "vaneckpm"])

    # Unlucky for me I know none of these.
    # def test_team_disbanded(self) -> None:
    #     """
    #     Test a challenge with a missing creator.
    #     """


if __name__ == '__main__':
    unittest.main()



from dataclasses import dataclass, field
from enum import Enum
from datetime import datetime
from typing import List, Dict, Union, Any
from ...utils.decorators import fetch_property
from ...exceptions import NotFound
from ..GameVersion import GameVersion
from bloonspy.utils.Infinity import Infinity
from ..Loadable import Loadable
from ..Asset import Asset
from .Restriction import Restriction, TowerRestriction
from .Gamemode import Gamemode
from .Power import Power
from .Tower import Tower
from .User import User


class ChallengeFilter(Enum):
    NEWEST = "newest"
    TRENDING = "trending"
    DAILY = "daily"


@dataclass(kw_only=True)
class ChallengeModifier:
    """All the modifiers of a challenge."""
    ability_cooldown_reduction: float = field(default=1.0)  #: Multiplier for ability cooldowns.
    removable_cost: float = field(default=1.0)  #: Multiplier for the cost of track removables.
    bloon_speed: float = field(default=1.0)  #: Bloon speed multiplier.
    moab_speed: float = field(default=1.0)  #: MOAB speed multiplier.
    boss_speed: float = field(default=1.0)  #: Boss bloon speed multiplier.
    ceramic_health: float = field(default=1.0)  #: Ceramic layer health multiplier.
    moab_health: float = field(default=1.0)  #: MOAB class bloon health multiplier.
    boss_health: float = field(default=1.0)  #: Boss bloon health multiplier
    regrow_rate: float = field(default=1.0)  #: Regrow rate multiplier
    all_regrow: bool = field(default=False)  #: If `True`, all bloons will be regrow.
    all_camo: bool = field(default=False)  #: If `True`, all bloons will be camo.


class Challenge(Loadable):
    """A BTD6 Challenge. It extends :class:`~bloonspy.model.Loadable`."""

    endpoint = "/btd6/challenges/challenge/{}"

    def __init__(self, challenge_id: str, eager: bool = True, name: str = None, created_at: int = None,
                 creator_id: str = None, raw_challenge: Dict[str, Any] = None):
        """Constructor method."""
        super().__init__(challenge_id, eager=eager)
        if raw_challenge:
            self._parse_json(raw_challenge)
        if name:
            self._data["name"] = name
        if created_at:
            self._data["createdAt"] = datetime.fromtimestamp(int(created_at/1000))
        if creator_id:
            self._data["creatorId"] = creator_id

    def _handle_exceptions(self, exception: Exception) -> None:
        error_msg = str(exception)
        if error_msg == "No challenge with that ID exists":
            raise NotFound(error_msg)

    def _parse_json(self, raw_challenge: Dict[str, Any]) -> None:
        self._loaded = False

        copy_keys = [
            "name", "disableDoubleCash", "disableInstas", "disableMK", "disablePowers", "disableSelling",
            "startingCash", "noContinues", "seed", "roundSets", "lives", "maxLives",
            "startRound", "endRound", "maxTowers", "maxParagons", "plays", "wins", "losses", "upvotes", "playsUnique",
            "winsUnique", "lossesUnique"
        ]
        for key in copy_keys:
            self._data[key] = raw_challenge[key]

        # self._data["maxTowers"] = Infinity() if raw_challenge["maxTowers"] == 9999 else raw_challenge["maxTowers"]
        # self._data["maxParagons"] = Infinity() if raw_challenge["maxParagons"] == 10 else raw_challenge["maxParagons"]
        self._data["leastCash"] = Infinity() if raw_challenge["leastCashUsed"] == -1 else raw_challenge["leastCashUsed"]
        self._data["leastTiers"] = Infinity() if raw_challenge["leastTiersUsed"] == -1 else raw_challenge["leastTiersUsed"]
        self._data["createdAt"] = datetime.fromtimestamp(int(raw_challenge["createdAt"]/1000))
        self._data["gameVersion"] = GameVersion.from_string(raw_challenge["gameVersion"])
        self._data["map"] = Asset(raw_challenge["map"], raw_challenge["mapURL"])
        self._data["gamemode"] = Gamemode.from_strings(raw_challenge["difficulty"], raw_challenge["mode"])
        self._data["creatorId"] = None
        if raw_challenge["creator"]:
            creator_id = raw_challenge["creator"].split("/")[-1]
            self._data["creatorId"] = creator_id

        self._data["powers"] = {}
        for power in raw_challenge["_powers"]:
            if power["max"] == 0:
                continue
            amount = Infinity() if power["max"] == -1 else power["max"]
            self._data["powers"][Power.from_string(power["power"])] = amount

        # Renames the keys so they can be fed into the dataclass
        modifiers = {}
        rename_keys = [
            ("abilityCooldownReductionMultiplier", "ability_cooldown_reduction"),
            ("removeableCostMultiplier", "removable_cost"),
            ("_bloonModifiers.speedMultiplier", "bloon_speed"),
            ("_bloonModifiers.moabSpeedMultiplier", "moab_speed"),
            ("_bloonModifiers.bossSpeedMultiplier", "boss_speed"),
            ("_bloonModifiers.healthMultipliers.bloons", "ceramic_health"),
            ("_bloonModifiers.healthMultipliers.moabs", "moab_health"),
            ("_bloonModifiers.healthMultipliers.boss", "boss_health"),
            ("_bloonModifiers.regrowRateMultiplier", "regrow_rate"),
            ("_bloonModifiers.allRegen", "all_regrow"),
            ("_bloonModifiers.allCamo", "all_camo"),
        ]
        for old_key, new_key in rename_keys:
            old_keys = old_key.split(".")
            item = raw_challenge
            while len(old_keys) > 0:
                if old_keys[0] not in item:
                    item = None
                    break
                item = item[old_keys[0]]
                old_keys.pop(0)

            if item is None:
                continue
            modifiers[new_key] = item
        self._data["modifiers"] = ChallengeModifier(**modifiers)

        self._data["towers"] = {}
        if raw_challenge["_towers"] is not None:  # Is null in Odysseys
            for restriction_data in raw_challenge["_towers"]:
                tower = Tower.from_string(restriction_data["tower"])
                max_towers = Infinity() if restriction_data["max"] == -1 else restriction_data["max"]
                if tower is None or max_towers == 0:
                    continue

                if tower.is_hero():
                    self._data["towers"][tower] = Restriction(max_towers=max_towers)
                else:
                    self._data["towers"][tower] = TowerRestriction(
                        max_towers=max_towers,
                        top_path_blocked=restriction_data["path1NumBlockedTiers"],
                        middle_path_blocked=restriction_data["path2NumBlockedTiers"],
                        bottom_path_blocked=restriction_data["path3NumBlockedTiers"]
                    )

        self._loaded = True

    @property
    def id(self) -> str:
        """The unique ID of the challenge."""
        return self._id

    @property
    @fetch_property(Loadable.load_resource, should_load=Loadable._should_load_property("name"))
    def name(self) -> str:
        """The name of the challenge."""
        return self._data["name"]

    @property
    @fetch_property(Loadable.load_resource, should_load=Loadable._should_load_property("createdAt"))
    def created_at(self) -> datetime:
        """The time the challenge was created at."""
        return self._data["createdAt"]

    @property
    @fetch_property(Loadable.load_resource)
    def game_version(self) -> GameVersion:
        """The latest game version the challenge was last beaten in."""
        return self._data["gameVersion"]

    @property
    @fetch_property(Loadable.load_resource, should_load=Loadable._should_load_property("creatorId"))
    def creator_id(self) -> str:
        """The ID of the challenge's creator. `None` if there isn't one (e.g. Odyssey challenges)."""
        return self._data["creatorId"]

    @fetch_property(Loadable.load_resource)
    def creator(self) -> User or None:
        """Fetch the creator of the challenge.

        .. warning::
           This function needs the property :attr:`~bloonspy.model.btd6.Challenge.creator_id` to be
           loaded, or it will make another API call to fetch that first.

        :return: The creator of the challenge of `None` if there isn't one.
        :rtype: User or None
        """
        if self.creator_id is None:
            return None
        return User(self.creator_id, eager=True)

    @property
    @fetch_property(Loadable.load_resource)
    def challenge_map(self) -> Asset:
        """Name and URL of the map the challenge takes place in."""
        return self._data["map"]

    @property
    @fetch_property(Loadable.load_resource)
    def gamemode(self) -> Gamemode:
        """Gamemode of the challenge."""
        return self._data["gamemode"]

    @property
    @fetch_property(Loadable.load_resource)
    def disable_double_cash(self) -> bool:
        """`True` if Double Cash is disabled."""
        return self._data["disableDoubleCash"]

    @property
    @fetch_property(Loadable.load_resource)
    def disable_instas(self) -> bool:
        """`True` if Insta Monkeys is disabled."""
        return self._data["disableInstas"]

    @property
    @fetch_property(Loadable.load_resource)
    def disable_monkey_knowledge(self) -> bool:
        """`True` if Monkey Knowledge is disabled."""
        return self._data["disableMK"]

    @property
    @fetch_property(Loadable.load_resource)
    def disable_powers(self) -> bool:
        """`True` if Powers are disabled."""
        return self._data["disablePowers"]

    @property
    @fetch_property(Loadable.load_resource)
    def disable_selling(self) -> bool:
        """`True` if Selling is disabled."""
        return self._data["disableSelling"]

    @property
    @fetch_property(Loadable.load_resource)
    def disable_continues(self) -> bool:
        """`True` if Continues are disabled."""
        return self._data["noContinues"]

    @property
    @fetch_property(Loadable.load_resource)
    def starting_cash(self) -> int:
        """The amount of cash you start with."""
        return self._data["startingCash"]

    @property
    @fetch_property(Loadable.load_resource)
    def starting_lives(self) -> int:
        """The amount of lives you start with."""
        return self._data["lives"]

    @property
    @fetch_property(Loadable.load_resource)
    def max_lives(self) -> int:
        """The maximum amount of lives you can have."""
        return self._data["maxLives"]

    @property
    @fetch_property(Loadable.load_resource)
    def max_towers(self) -> int:
        """The maximum amount of towers you can have at any given time."""
        return self._data["maxTowers"]

    @property
    @fetch_property(Loadable.load_resource)
    def max_paragons(self) -> int:
        """The maximum amount of paragons you can have at any given time."""
        return self._data["maxParagons"]

    @property
    @fetch_property(Loadable.load_resource)
    def start_round(self) -> int:
        """The round the challenge starts at."""
        return self._data["startRound"]

    @property
    @fetch_property(Loadable.load_resource)
    def end_round(self) -> int:
        """The round the challenge ends at."""
        return self._data["endRound"]

    @property
    @fetch_property(Loadable.load_resource)
    def plays(self) -> int:
        """How many times the challenge has been played."""
        return self._data["plays"]

    @property
    @fetch_property(Loadable.load_resource)
    def plays_unique(self) -> int:
        """Amount of unique people that have played the challenge."""
        return self._data["playsUnique"]

    @property
    @fetch_property(Loadable.load_resource)
    def wins(self) -> int:
        """How many times the challenge has been won."""
        return self._data["winsUnique"]

    @property
    @fetch_property(Loadable.load_resource)
    def wins_unique(self) -> int:
        """Amount of unique people that have won the challenge."""
        return self._data["winsUnique"]

    @property
    @fetch_property(Loadable.load_resource)
    def losses(self) -> int:
        """How many times the challenge has been lost."""
        return self._data["losses"]

    @property
    @fetch_property(Loadable.load_resource)
    def losses_unique(self) -> int:
        """Amount of unique people that have lost the challenge."""
        return self._data["lossesUnique"]

    @property
    @fetch_property(Loadable.load_resource)
    def upvotes(self) -> int:
        """Amount of upvotes the challenge has."""
        return self._data["upvotes"]

    @property
    @fetch_property(Loadable.load_resource)
    def least_cash_used(self) -> Union[int, Infinity]:
        """Least Cash restriction on the challenge. If there's none, it's :class:`Infinity`."""
        return self._data["leastCash"]

    @property
    @fetch_property(Loadable.load_resource)
    def least_tiers_used(self) -> Union[int, Infinity]:
        """Least Tiers restriction on the challenge. If there's none, it's :class:`Infinity`."""
        return self._data["leastTiers"]

    @property
    @fetch_property(Loadable.load_resource)
    def seed(self) -> int:
        """The RNG seed for the challenge."""
        return self._data["seed"]

    @property
    @fetch_property(Loadable.load_resource)
    def round_sets(self) -> List[str]:
        """Names of the round sets of the challenge."""
        return self._data["roundSets"]

    @property
    @fetch_property(Loadable.load_resource)
    def powers(self) -> Dict[Power, Union[int, Infinity]]:
        """The powers allowed for the challenge, and how many you can use of each."""
        return self._data["powers"]

    @property
    @fetch_property(Loadable.load_resource)
    def modifiers(self) -> Dict[ChallengeModifier, float]:
        """Challenge modifiers, such as bloon speeds and health."""
        return self._data["modifiers"]

    @property
    @fetch_property(Loadable.load_resource)
    def towers(self) -> Dict[Tower, Restriction]:
        """Tower and Hero restrictions on the challenge."""
        return self._data["towers"]



from dataclasses import dataclass
from .Tower import Tower


@dataclass
class InstaMonkey:
    """An Insta Monkey."""
    tower: Tower  #: The tower this Insta Monkey is for.
    top_path: int  #: Top path upgrades.
    middle_path: int  #: Middle path upgrades.
    bottom_path: int  #: Bottom path upgrades.


@dataclass
class Reward:
    """A generic reward."""
    type: str  #: Name of the reward, e.g. `MonkeyMoney`.
    amount: int  #: Amount of the reward.



from dataclasses import dataclass
from typing import Union
from ...utils.Infinity import Infinity


@dataclass(kw_only=True)
class Restriction:
    """Generic tower restriction."""
    max_towers: Union[int, Infinity]  #: Max amount of towers you can place of a kind.


@dataclass(kw_only=True)
class TowerRestriction(Restriction):
    """Non-hero restriction. Inherits from :class:`~bloonspy.model.btd6.Restriction`."""
    top_path_blocked: int  #: Number of restricted top path upgrades.
    middle_path_blocked: int  #: Number of restricted middle path upgrades.
    bottom_path_blocked: int  #: Number of restricted bottom path upgrades.



from dataclasses import dataclass
from typing import Union


@dataclass
class GameVersion:
    major: int  #: The major version of the game.
    minor: int  #: The minor version of the game.

    @staticmethod
    def from_string(version: Union[str, int]) -> "GameVersion":
        if version == 0:
            return GameVersion(0, 0)

        major = minor = 0
        versions = version.split(".")
        if len(versions) > 0 and versions[0].isnumeric():
            major = int(versions[0])
        if len(versions) > 1 and versions[1].isnumeric():
            minor = int(versions[1])

        return GameVersion(major, minor)



from typing import Dict, Any
from ..utils.api import get
from ..utils.decorators import exception_handler


class Loadable:
    """Represents a resource that can be loaded."""
    endpoint = "{}"

    def __init__(self, resource_id: str, eager: bool = True):
        self._id = resource_id
        self._data = {}
        self._loaded = False
        if eager:
            self.load_resource()

    def handle_exceptions(self, exception: Exception) -> None:
        return self._handle_exceptions(exception)

    def _handle_exceptions(self, exception: Exception) -> None:
        raise exception

    @exception_handler(handle_exceptions)
    def load_resource(self, only_if_unloaded: bool = True) -> None:
        """Load the resource.

        :param only_if_unloaded: Only make an API call if the resource is unloaded.
            If `False`, it essentially "reloads" the resource.
        :type only_if_unloaded: :class:`bool`

        :raises bloonspy.exceptions.NotFound: If the resource is not found.
        """
        if self._loaded and only_if_unloaded:
            return

        data = get(self.endpoint.format(self._id))
        self._parse_json(data)

    @staticmethod
    def _should_load_property(key_name: str) -> callable:
        def _inner(self: Loadable) -> bool:
            return key_name not in self._data
        return _inner

    def _parse_json(self, raw_user: Dict[str, Any]) -> None:
        self._loaded = True

    @property
    def id(self) -> str:
        """The unique ID of the resource."""
        return self._id

    @property
    def loaded(self) -> bool:
        """`True` if the resource is loaded."""
        return self._loaded



from dataclasses import dataclass, field
from typing import Dict, Any
from ...exceptions import NotFound
from ...utils.decorators import fetch_property
from ...utils.dictionaries import rename_keys
from ..Loadable import Loadable
from ..Asset import Asset
from .Tower import Tower
from .Medals import EventMedals, MapMedals, CTGlobalMedals, CTLocalMedals


@dataclass(kw_only=True)
class BloonsPoppedStats:
    """How many bloons a user has popped of each type."""
    total: int = field(default=0)  #: Total bloons popped.
    total_coop: int = field(default=0)  #: Total bloons popped in coop games.
    camos: int = field(default=0)  #: Number of camo bloons popped.
    regrows: int = field(default=0)  #: Number of regrow bloons popped.
    purples: int = field(default=0)  #: Number of purple bloons popped.
    leads: int = field(default=0)  #: Number of lead bloons popped.
    ceramics: int = field(default=0)  #: Number of ceramic bloons popped.
    moabs: int = field(default=0)  #: Number of MOAB bloons popped.
    bfbs: int = field(default=0)  #: Number of BFB bloons popped.
    zomgs: int = field(default=0)  #: Number of ZOMG bloons popped.
    bads: int = field(default=0)  #: Number of BAD bloons popped.
    golden: int = field(default=0)  #: Number of golden bloons popped.


@dataclass(kw_only=True)
class GameplayStats:
    """User's gameplay stats."""
    most_experienced_monkey: Tower = field(default=Tower.DART_MONKEY)  #: Monkey with the most XP.
    cash_earned: int = field(default=0)  #: Total cash earned.
    challenges_completed: int = field(default=0)  #: Total challenges completed.
    collection_chests_opened: int = field(default=0)  #: Number of collection chests opened.
    coop_cash_given: int = field(default=0)  #: Total cash gifted in coop games.
    daily_rewards: int = field(default=0)  #: Total daily rewards collected.
    game_count: int = field(default=0)  #: Number of games played
    games_won: int = field(default=0)  #: Number of games won
    highest_round: int = field(default=0)  #: Highest round beaten
    highest_round_chimps: int = field(default=0)  #: Highest round beaten in CHIMPS mode
    highest_round_deflation: int = field(default=0)  #: Highest round beaten in Deflation mode
    insta_monkey_collection: int = field(default=0)  #: Number of unique Insta Monkeys collected
    monkey_teams_wins: int = field(default=0)  #: Number of games won with the Monkey Teams restrictions.
    powers_used: int = field(default=0)  #: Number of powers used.
    total_odysseys_completed: int = field(default=0)  #: Number of odysseys completed.
    total_odyssey_stars: int = field(default=0)  #: Total odyssey stars.
    total_trophies_earned: int = field(default=0)  #: Lifetime trophies earned through events.
    necro_bloons_reanimated: int = field(default=0)  #: Number of necro bloons reanimated.
    bloons_leaked: int = field(default=0)  #: Total RBE leaked.
    bloons_popped: BloonsPoppedStats = field(default_factory=BloonsPoppedStats)  #: In depth stats about bloons popped.


class User(Loadable):
    """A BTD6 player. Inherits from :class:`~bloonspy.model.Loadable`."""
    endpoint = "/btd6/users/{}"

    def _handle_exceptions(self, exception: Exception) -> None:
        error_msg = str(exception)
        if error_msg == "Invalid user ID / Player Does not play this game":
            raise NotFound(error_msg)

    def _parse_json(self, raw_user: Dict[str, Any]) -> None:
        self._loaded = False

        copy_keys = ["displayName", "rank", "veteranRank", "achievements", "followers"]
        for key in copy_keys:
            self._data[key] = raw_user[key]

        self._data["avatar"] = Asset(raw_user["avatar"], raw_user["avatarURL"])
        self._data["banner"] = Asset(raw_user["banner"], raw_user["bannerURL"])

        map_medal_keys = [
            ("CHIMPS-BLACK", "chimps_black"), ("Clicks", "chimps_red"), ("Easy", "easy"), ("Medium", "medium"),
            ("Hard", "hard"), ("PrimaryOnly", "primary_only"), ("Deflation", "deflation"),
            ("MilitaryOnly", "military_only"), ("Apopalypse", "apopalypse"), ("Reverse", "reverse"),
            ("MagicOnly", "magic_only"), ("HalfCash", "half_cash"), ("DoubleMoabHealth", "double_hp_moabs"),
            ("AlternateBloonsRounds", "alternate_bloons_rounds"), ("Impoppable", "impoppable")
        ]
        self._data["single_player_medals"] = MapMedals(
            **rename_keys(raw_user["_medalsSinglePlayer"], map_medal_keys)
        )
        self._data["coop_medals"] = MapMedals(
            **rename_keys(raw_user["_medalsMultiplayer"], map_medal_keys)
        )

        event_medal_keys = [
            ("BlackDiamond", "first"), ("RedDiamond", "second"), ("Diamond", "third"), ("GoldDiamond", "top_50"),
            ("DoubleGold", "top_1_percent"), ("GoldSilver", "top_10_percent"), ("DoubleSilver", "top_25_percent"),
            ("Silver", "top_50_percent"), ("Bronze", "top_75_percent")
        ]
        self._data["boss_normal_medals"] = EventMedals(
            **rename_keys(raw_user["_medalsBoss"], event_medal_keys)
        )
        self._data["boss_elite_medals"] = EventMedals(
            **rename_keys(raw_user["_medalsBossElite"], event_medal_keys)
        )
        # self._data["race_medals"] = Medals(
        #     **rename_keys(raw_user["_medals"], event_medal_keys)
        # )

        ct_local_medal_keys = [
            ("BlackDiamond", "first"), ("RedDiamond", "second"), ("Diamond", "third"), ("GoldDiamond", "top_10"),
            ("DoubleGold", "top_20"), ("Silver", "top_40"), ("Bronze", "top_60")
        ]
        ct_global_medal_keys = [
            ("Diamond", "top_25"), ("GoldDiamond", "top_100"), ("DoubleGold", "top_1_percent"),
            ("GoldSilver", "top_10_percent"), ("DoubleSilver", "top_25_percent"), ("Silver", "top_50_percent"),
            ("Bronze", "top_75_percent")
        ]
        self._data["ct_local_medals"] = CTLocalMedals(
            **rename_keys(raw_user["_medalsCTLocal"], ct_local_medal_keys)
        )
        self._data["ct_global_medals"] = CTGlobalMedals(
            **rename_keys(raw_user["_medalsCTLocal"], ct_global_medal_keys)
        )

        stats_keys = [
            ("gameplay.cashEarned", "cash_earned"),
            ("gameplay.challengesCompleted", "challenges_completed"),
            ("gameplay.collectionChestsOpened", "collection_chests_opened"),
            ("gameplay.coopCashGiven", "coop_cash_given"),
            ("gameplay.dailyRewards", "daily_rewards"),
            ("gameplay.gameCount", "game_count"),
            ("gameplay.gamesWon", "games_won"),
            ("gameplay.highestRound", "highest_round"),
            ("gameplay.highestRoundCHIMPS", "highest_round_chimps"),
            ("gameplay.highestRoundDeflation", "highest_round_deflation"),
            ("gameplay.instaMonkeyCollection", "insta_monkey_collection"),
            ("gameplay.monkeyTeamsWins", "monkey_teams_wins"),
            ("gameplay.powersUsed", "powers_used"),
            ("gameplay.totalOdysseysCompleted", "total_odysseys_completed"),
            ("gameplay.totalOdysseyStars", "total_odyssey_stars"),
            ("gameplay.totalTrophiesEarned", "total_trophies_earned"),
            ("bloonsPopped.necroBloonsReanimated", "necro_bloons_reanimated"),
            ("bloonsPopped.bloonsLeaked", "bloons_leaked"),
        ]
        bloons_popped_keys = [
            ("badsPopped", "bads"),
            ("bfbsPopped", "bfbs"),
            ("bloonsPopped", "total"),
            ("camosPopped", "camos"),
            ("ceramicsPopped", "ceramics"),
            ("coopBloonsPopped", "total_coop"),
            ("goldenBloonsPopped", "golden"),
            ("leadsPopped", "leads"),
            ("moabsPopped", "moabs"),
            ("purplesPopped", "purples"),
            ("regrowsPopped", "regrows"),
            ("zomgsPopped", "zomgs"),
        ]
        self._data["stats"] = GameplayStats(
            most_experienced_monkey=Tower.from_string(raw_user["mostExperiencedMonkey"]),
            bloons_popped=BloonsPoppedStats(**rename_keys(raw_user["bloonsPopped"], bloons_popped_keys)),
            **rename_keys(raw_user, stats_keys)
        )

        heroes_placed = {}
        for hero in raw_user["heroesPlaced"].keys():
            heroes_placed[Tower.from_string(hero)] = raw_user["heroesPlaced"][hero]
        self._data["heroes_placed"] = heroes_placed

        self._loaded = True

    @property
    @fetch_property(Loadable.load_resource)
    def name(self) -> str:
        """The name of the user."""
        return self._data["displayName"]

    @property
    @fetch_property(Loadable.load_resource)
    def rank(self) -> int:
        """The user's rank."""
        return self._data["rank"]

    @property
    @fetch_property(Loadable.load_resource)
    def veteran_rank(self) -> int:
        """The user's veteran rank, or 0 if they haven't reached it yet."""
        return self._data["veteranRank"]

    @property
    @fetch_property(Loadable.load_resource)
    def achievements(self) -> int:
        """Number of achievements this user unlocked."""
        return self._data["achievements"]

    @property
    @fetch_property(Loadable.load_resource)
    def followers(self) -> int:
        """Number of followers this user has."""
        return self._data["followers"]

    @property
    @fetch_property(Loadable.load_resource)
    def avatar(self) -> Asset:
        """The user's current avatar."""
        return self._data["avatar"]

    @property
    @fetch_property(Loadable.load_resource)
    def banner(self) -> Asset:
        """The user's current banner."""
        return self._data["banner"]

    @property
    @fetch_property(Loadable.load_resource)
    def single_player_medals(self) -> MapMedals:
        """Medals earned in single player mode."""
        return self._data["single_player_medals"]

    @property
    @fetch_property(Loadable.load_resource)
    def coop_medals(self) -> MapMedals:
        """Medals earned in coop mode."""
        return self._data["coop_medals"]

    @property
    @fetch_property(Loadable.load_resource)
    def boss_normal_medals(self) -> EventMedals:
        """Ranked normal boss mdeals."""
        return self._data["boss_normal_medals"]

    @property
    @fetch_property(Loadable.load_resource)
    def boss_elite_medals(self) -> EventMedals:
        """Ranked elite boss medals."""
        return self._data["boss_elite_medals"]

    # @property
    # @fetch_property(Loadable.load_resource)
    # def race_medals(self) -> EventMedals:
    #     """Race event medals."""
    #     return self._data["race_medals"]

    @property
    @fetch_property(Loadable.load_resource)
    def ct_local_medals(self) -> CTLocalMedals:
        """Contested Territory local medals."""
        return self._data["ct_local_medals"]

    @property
    @fetch_property(Loadable.load_resource)
    def ct_global_medals(self) -> CTGlobalMedals:
        """Contested Territory global medals."""
        return self._data["ct_global_medals"]

    @property
    @fetch_property(Loadable.load_resource)
    def stats(self) -> GameplayStats:
        """Gameplay stats."""
        return self._data["stats"]

    @property
    @fetch_property(Loadable.load_resource)
    def heroes_placed(self) -> Dict[Tower, int]:
        """Number of times each hero has been placed."""
        return self._data["heroes_placed"]



import unittest
from bloonspy import Client


class TestChallengeCreator(unittest.TestCase):
    def test_challenge(self) -> None:
        """
        Test a challenge with a creator.
        """
        challenge = Client.get_challenge("ZMDCDTB")
        creator = challenge.creator()
        self.assertEqual(creator.name, "Sarto")
        self.assertEqual(creator.id, "9cee138c8c94ffac1910864c0b73e577ca554ce8cb18db6c")

    def test_challenge_no_creator(self) -> None:
        """
        Test a challenge with a missing creator.
        """
        challenge = Client.get_challenge("rot168220230320")
        creator = challenge.creator()
        self.assertIsNone(creator)


if __name__ == '__main__':
    unittest.main()





class Infinity:
    """Class to represent infinity.

    To check whether something is infinity, just see if they're equal. ::

       not_infinity = 3
       print(not_infinity == Infinity)  # False
       infinity = Infinity()
       print(infinity == Infinity)  # True

    """
    def __str__(self):
        return "Inf"

    def __repr__(self):
        return "Inf"

    def __eq__(self, other):
        return isinstance(other, Infinity) or other == Infinity



import unittest
from datetime import datetime
import requests
from bloonspy import btd6


class TestRace(unittest.TestCase):
    def test_race(self) -> None:
        """
        Test that a race is loaded correctly.
        """
        races = requests.get("https://data.ninjakiwi.com/btd6/races")
        race_id = races.json()["body"][0]["id"]
        race = btd6.Race(race_id)

        check_instance = [
            ("name", str), ("start", datetime), ("end", datetime), ("total_scores", int),
            ("total_scores", int), ("gamemode", btd6.Gamemode)
        ]
        for attr_name, attr_type in check_instance:
            self.assertIsInstance(getattr(race, attr_name), attr_type,
                                  msg=f"Assert if Race.{attr_name} is {attr_type}")

    def test_race_expired(self) -> None:
        """
        Test that the correct exception is raised when attempting to fetch a nonexistant/expired race.
        """
        race_id = "Uncracked_lctxs1bb"
        correct_exception = False
        try:
            btd6.Race(race_id, eager=True)
        except btd6.NotFound:
            correct_exception = True
        self.assertTrue(correct_exception, msg="Wrong/expired race IDs should raise bloonspy.exceptions.NotFound")


if __name__ == '__main__':
    unittest.main()



from datetime import datetime
from typing import List, Dict, Any
from concurrent.futures import ThreadPoolExecutor
from ...utils.decorators import fetch_property, exception_handler
from ...utils.api import get, get_lb_page
from ..Event import Event
from .User import User
from .Team import Team


class CtPlayer(User):
    """A player who has played a CT event and is on the leaderboard.
    Inherits from :class:`~bloonspy.model.btd6.User`.
    """
    def __init__(self, user_id: str, name: str, score: int, **kwargs):
        super().__init__(user_id, **kwargs)
        self._name = name
        self._score = score

    @property
    def name(self) -> str:
        """The name of the user."""
        return self._name

    @property
    def score(self) -> int:
        """The current CT points of the user."""
        return self._score


class CtTeam(Team):
    """A team who participated in Contested Territory and is on the leaderboards."""
    def __init__(self, team_id: str, name: str, score: int, **kwargs):
        super().__init__(team_id, **kwargs)
        self._name = name
        self._score = score

    @property
    def name(self) -> str:
        """The name of the team."""
        return self._name

    @property
    def score(self) -> int:
        """The current total CT points of the team."""
        return self._score


class ContestedTerritoryEvent(Event):
    """A Contested Territory event. Inherits from :class:`~bloonspy.model.Event`."""
    event_endpoint = "/btd6/ct"
    lb_endpoint_player = "/btd6/ct/{}/leaderboard/player"
    lb_endpoint_team = "/btd6/ct/{}/leaderboard/team"
    event_dict_keys = ["name", "start", "end", "totalScores_player", "totalScores_team"]
    event_name = "CT"

    def _parse_event(self, data: Dict[str, Any]) -> None:
        self._data["totalScores_player"] = data["totalScores_player"]
        self._data["totalScores_team"] = data["totalScores_team"]
        self._data["start"] = datetime.fromtimestamp(data["start"]/1000)
        self._data["end"] = datetime.fromtimestamp(data["end"]/1000)
        self._event_loaded = True

    @property
    def name(self) -> str:
        return f"Contested Territory #{self.event_number}"

    @property
    @fetch_property(Event.load_event, should_load=Event._should_load_property)
    def event_number(self) -> int:
        """Number of the event."""
        first_event_start = datetime.strptime('2022-08-09 22', '%Y-%m-%d %H')
        return int((self.start-first_event_start).days/14) + 1

    @property
    @fetch_property(Event.load_event, should_load=Event._should_load_property)
    def total_scores_player(self) -> int:
        """Number of players who participated in the event."""
        return self._data["totalScores_player"]

    @property
    @fetch_property(Event.load_event, should_load=Event._should_load_property)
    def total_scores_team(self) -> int:
        """Number of teams who participated in the event."""
        return self._data["totalScores_team"]

    @exception_handler(Event.handle_exceptions)
    def leaderboard_player(self, pages: int = 1, start_from_page: int = 0) -> List[CtPlayer]:
        """Get a page of the player leaderboard.

        .. note::
           The returned :class:`~bloonspy.model.btd6.CtPlayer` objects will only
           have the properties :attr:`~bloonspy.model.Loadable.id`, :attr:`~bloonspy.model.btd6.User.name`, and
           :attr:`~bloonspy.model.CtPlayer.score` loaded.

        :param pages: Number of pages to fetch.
        :type pages: int
        :param start_from_page: The first page to fetch.
        :type start_from_page: int

        :return: A list of players in the leaderboard.
        :rtype: List[:class:`~bloonspy.model.btd6.CtPlayer`]

        :raise ~bloonspy.exceptions.NotFound: If the boss doesn't exist or is expired.
        """
        futures = []
        with ThreadPoolExecutor(max_workers=10) as executor:
            for page_num in range(start_from_page, start_from_page + pages):
                futures.append(executor.submit(get_lb_page, self.lb_endpoint_player.format(self._id), page_num))

        players = []
        for page in futures:
            for player in page.result():
                players.append(CtPlayer(
                    player["profile"].split("/")[-1], player["displayName"], player["score"]
                ))

        return players

    @exception_handler(Event.handle_exceptions)
    def leaderboard_team(self, pages: int = 1, start_from_page: int = 0) -> List[CtTeam]:
        """Get a page of the team leaderboard.

        .. note::
           The returned :class:`~bloonspy.model.btd6.CtTeam` objects will only
           have the properties :attr:`~bloonspy.model.Loadable.id`, :attr:`~bloonspy.model.btd6.Team.name`,
           and :attr:`~bloonspy.model.CtTeam.score` loaded.

        :param pages: Number of pages to fetch.
        :type pages: int
        :param start_from_page: The first page to fetch.
        :type start_from_page: int

        :return: A list of teams in the leaderboard.
        :rtype: List[:class:`~bloonspy.model.btd6.CtTeam`]

        :raise ~bloonspy.exceptions.NotFound: If the boss doesn't exist or is expired.
        """
        futures = []
        with ThreadPoolExecutor(max_workers=10) as executor:
            for page_num in range(start_from_page, start_from_page + pages):
                futures.append(executor.submit(get_lb_page, self.lb_endpoint_team.format(self._id), page_num))

        teams = []
        for page in futures:
            for team in page.result():
                teams.append(CtTeam(
                    team["profile"].split("/")[-1], team["displayName"], team["score"]
                ))

        return teams



import unittest
from bloonspy import btd6, Client


class TestFetchTeamOwner(unittest.TestCase):
    def test_team_owner(self) -> None:
        """
        Test getting a team's owner.
        """
        user_id = "9cee138c8c94ffac1910864c0b73e577ca554ce8cb18db6c"
        user = Client.get_user(user_id)

        check_instance = [
            ("followers", int),
            ("name", str),
            ("banner", btd6.Asset),
            ("single_player_medals", btd6.MapMedals),
            ("boss_normal_medals", btd6.EventMedals),
            ("ct_local_medals", btd6.CTLocalMedals),
            ("stats", btd6.GameplayStats),
        ]
        for attr_name, attr_type in check_instance:
            self.assertIsInstance(getattr(user, attr_name), attr_type,
                                  msg=f"Assert if challenge.{attr_name} is {attr_type}")


if __name__ == '__main__':
    unittest.main()



from enum import Enum
from typing import List, Dict, Any, Union
from ...utils.decorators import fetch_property, exception_handler
from ...utils.Infinity import Infinity
from ...utils.api import get
from ..Event import Event
from ..Loadable import Loadable
from .Restriction import Restriction, TowerRestriction
from .Rewards import InstaMonkey, Reward
from .Challenge import Challenge
from .Power import Power
from .Tower import Tower


class OdysseyDifficulty(Enum):
    EASY = "easy"
    MEDIUM = "medium"
    HARD = "hard"


class Odyssey(Loadable):
    """Represents a single Odyssey. Inherits from :class:`~bloonspy.model.Loadable`."""
    endpoint = "/btd6/odyssey/{}/:difficulty:"
    map_endpoint = "/btd6/odyssey/{}/:difficulty:/maps"

    def __init__(self, resource_id: str, name: str, difficulty: OdysseyDifficulty, eager: bool = True):
        self.endpoint = self.endpoint.replace(":difficulty:", difficulty.value)
        self.map_endpoint = self.map_endpoint.replace(":difficulty:", difficulty.value)
        super().__init__(resource_id, eager)
        self._name = name
        self._difficulty = difficulty

    def _parse_json(self, raw_odyssey: Dict[str, Any]) -> None:
        self._loaded = False

        copy_keys = [
            "isExtreme", "maxMonkeySeats", "maxMonkeysOnBoat", "maxPowerSlots", "startingHealth",
        ]
        for key in copy_keys:
            self._data[key] = raw_odyssey[key]

        self._data["rewards"] = []
        for raw_reward in raw_odyssey["_rewards"]:
            reward_type, reward = raw_reward.split(":", 1)
            if reward_type == "InstaMonkey":
                tower, path = reward.split(",")
                path = int(path)
                self._data["rewards"].append(InstaMonkey(
                    Tower.from_string(tower), int(path/100), int(path/10) % 10, path % 10
                ))
            elif reward_type == "Power":
                self._data["rewards"].append(Power.from_string(reward))
            else:
                self._data["rewards"].append(Reward(reward_type, reward))
        self._data["rewards"] = raw_odyssey["_rewards"]

        self._data["availablePowers"] = {}
        for power in raw_odyssey["_availablePowers"]:
            if power["max"] == 0:
                continue
            self._data["availablePowers"][Power.from_string(power["power"])] = power["max"]

        self._data["defaultPowers"] = {}
        for power in raw_odyssey["_defaultPowers"]:
            self._data["defaultPowers"][Power.from_string(power["name"])] = power["quantity"]

        self._data["availableTowers"] = {}
        for tower in raw_odyssey["_availableTowers"]:
            if tower["isHero"]:
                restr = Restriction(max_towers=1)
            else:
                restr = TowerRestriction(
                    max_towers=Infinity(),
                    top_path_blocked=tower["path1NumBlockedTiers"],
                    middle_path_blocked=tower["path2NumBlockedTiers"],
                    bottom_path_blocked=tower["path3NumBlockedTiers"],
                )
            self._data["availableTowers"][Tower.from_string(tower["tower"])] = restr

        self._data["defaultTowers"] = {}
        for tower in raw_odyssey["_defaultTowers"]:
            self._data["defaultTowers"][Tower.from_string(tower["name"])] = tower["quantity"]

        self._loaded = True

    @property
    def name(self) -> str:
        """The Odyssey's name"""
        return self._name

    @property
    def difficulty(self) -> OdysseyDifficulty:
        """The Odysseys's difficulty"""
        return self._difficulty

    @property
    @fetch_property(Loadable.load_resource)
    def is_extreme(self) -> bool:
        """`True` if the odyssey is extreme."""
        return self._data["isExtreme"]

    @property
    @fetch_property(Loadable.load_resource)
    def max_monkey_seats(self) -> int:
        """Max number of monkey types you can take."""
        return self._data["maxMonkeySeats"]

    @property
    @fetch_property(Loadable.load_resource)
    def max_boat_seats(self) -> int:
        """Max number of monkeys you can take."""
        return self._data["maxMonkeysOnBoat"]

    @property
    @fetch_property(Loadable.load_resource)
    def max_power_slots(self) -> int:
        """Max number of powers you can take."""
        return self._data["maxPowerSlots"]

    @property
    @fetch_property(Loadable.load_resource)
    def starting_lives(self) -> int:
        """Odyssey's starting lives"""
        return self._data["startingHealth"]

    @property
    @fetch_property(Loadable.load_resource)
    def rewards(self) -> List[Union[Power, InstaMonkey, Reward]]:
        """Rewards for completing the odyssey."""
        return self._data["rewards"]

    @property
    def available_powers(self) -> Dict[Power, int]:
        """Available powers to choose from for the odyssey."""
        return self._data["availablePowers"]

    @property
    def default_powers(self) -> Dict[Power, int]:
        """Default powers for the odyssey."""
        return self._data["defaultPowers"]

    @property
    def available_towers(self) -> Dict[Tower, Restriction]:
        """Available towers to choose from for the odyssey."""
        return self._data["availableTowers"]

    @property
    def default_towers(self) -> Dict[Tower, int]:
        """Default towers for the odyssey."""
        return self._data["defaultTowers"]

    @exception_handler(Loadable.handle_exceptions)
    def maps(self) -> List[Challenge]:
        """Get all of the odyssey's challenges."""
        odyssey_map = get(self.map_endpoint.format(self._id))
        islands = []
        for island in odyssey_map:
            islands.append(Challenge(island["id"], raw_challenge=island))
        return islands


class OdysseyEvent(Event):
    """An Odyssey event. Inherits from :class:`~bloonspy.model.Event`."""
    event_endpoint = "/btd6/odyssey"
    event_name = "Odyssey"

    def easy(self, eager: bool = False) -> Odyssey:
        """Get the easy mode version of the Odyssey.

        .. note::
           If lazily loaded, the returned :class:`~bloonspy.model.btd6.Odyssey` object will only
           have the properties :attr:`~bloonspy.model.Loadable.id`, :attr:`~bloonspy.model.btd6.Odyssey.name`,
           and :attr:`~bloonspy.model.Odyssey.difficulty` loaded.

        :param eager: If `True`, it loads all of the data right away. Set it to `False`
            if you want to limit API calls and don't need all the data. For more information,
            please read `Lazy and Eager Loading <async.html#lazy-and-eager-loading>`_.
        :type eager: bool
        :return: The easy mode of the odyssey.
        :rtype: ~bloonspy.model.btd6.Odyssey
        """
        return Odyssey(self.id, self.name, OdysseyDifficulty.EASY, eager=eager)

    def medium(self, eager: bool = False) -> Odyssey:
        """Get the medium mode version of the Odyssey.

        .. note::
           If lazily loaded, the returned :class:`~bloonspy.model.btd6.Odyssey` object will only
           have the properties :attr:`~bloonspy.model.Loadable.id`, :attr:`~bloonspy.model.btd6.Odyssey.name`,
           and :attr:`~bloonspy.model.Odyssey.difficulty` loaded.

        :param eager: If `True`, it loads all of the data right away. Set it to `False`
            if you want to limit API calls and don't need all the data. For more information,
            please read `Lazy and Eager Loading <async.html#lazy-and-eager-loading>`_.
        :type eager: bool
        :return: The medium mode of the odyssey.
        :rtype: ~bloonspy.model.btd6.Odyssey
        """
        return Odyssey(self.id, self.name, OdysseyDifficulty.EASY, eager=eager)

    def hard(self, eager: bool = False) -> Odyssey:
        """Get the hard mode version of the Odyssey.

        .. note::
           If lazily loaded, the returned :class:`~bloonspy.model.btd6.Odyssey` object will only
           have the properties :attr:`~bloonspy.model.Loadable.id`, :attr:`~bloonspy.model.btd6.Odyssey.name`,
           and :attr:`~bloonspy.model.Odyssey.difficulty` loaded.

        :param eager: If `True`, it loads all of the data right away. Set it to `False`
            if you want to limit API calls and don't need all the data. For more information,
            please read `Lazy and Eager Loading <async.html#lazy-and-eager-loading>`_.
        :type eager: bool
        :return: The hard mode of the odyssey.
        :rtype: ~bloonspy.model.btd6.Odyssey
        """
        return Odyssey(self.id, self.name, OdysseyDifficulty.EASY, eager=eager)



import unittest
from datetime import datetime
import bloonspy
from bloonspy import btd6


class TestUser(unittest.TestCase):
    def test_user(self) -> None:
        """
        Test that an user is loaded correctly.
        """
        user_id = "9cee138c8c94ffac1910864c0b73e577ca554ce8cb18db6c"
        user = btd6.User(user_id)

        expected_results = [
            ("id", "9cee138c8c94ffac1910864c0b73e577ca554ce8cb18db6c"),
            ("name", "Sarto"),
            ("rank", 155),
        ]
        for attr_name, attr_expected_value in expected_results:
            self.assertEqual(getattr(user, attr_name), attr_expected_value,
                msg=f"Asserting User.{attr_name}")

        # Greater or Equal cause these can increase in the future.
        expected_results_ge = [
            ("achievements", 124),
            ("veteran_rank", 8),
        ]
        for attr_name, attr_expected_min in expected_results_ge:
            self.assertGreaterEqual(getattr(user, attr_name), attr_expected_min,
                msg=f"Asserting User.{attr_name}")

        check_instance = [
            ("followers", int),
            ("avatar", btd6.Asset),
            ("banner", btd6.Asset),
            ("single_player_medals", btd6.MapMedals),
            ("coop_medals", btd6.MapMedals),
            ("boss_normal_medals", btd6.EventMedals),
            ("boss_elite_medals", btd6.EventMedals),
            # ("race_medals", btd6.Medals),
            ("ct_local_medals", btd6.CTLocalMedals),
            ("ct_global_medals", btd6.CTGlobalMedals),
            ("stats", btd6.GameplayStats),
        ]
        for attr_name, attr_type in check_instance:
            self.assertIsInstance(getattr(user, attr_name), attr_type,
                msg=f"Assert if challenge.{attr_name} is {attr_type}")

        # Check medals

        expected_results_ge = [
            ("chimps_black", 1), ("easy", 1), ("medium", 1), ("hard", 1),
            ("primary_only", 1), ("deflation", 1), ("military_only", 1),
            ("apopalypse", 1), ("reverse", 1), ("magic_only", 1), ("half_cash", 1),
            ("double_hp_moabs", 1), ("alternate_bloons_rounds", 1), ("impoppable", 1),
            ("chimps_red", 1),
        ]
        for attr_name, attr_expected_min in expected_results_ge:
            self.assertGreaterEqual(getattr(user.single_player_medals, attr_name), attr_expected_min,
                                    msg=f"Asserting User.single_player_medals.{attr_name}")
            self.assertGreaterEqual(getattr(user.coop_medals, attr_name), attr_expected_min,
                                    msg=f"Asserting User.coop_medals.{attr_name}")

        check_instance = [
            ("first", int), ("second", int), ("third", int), ("top_50", int), ("top_1_percent", int),
            ("top_10_percent", int), ("top_25_percent", int), ("top_50_percent", int), ("top_75_percent", int)
        ]
        for attr_name, attr_type in check_instance:
            self.assertIsInstance(getattr(user.boss_normal_medals, attr_name), attr_type,
                                  msg=f"Asserting User.boss_normal_medals.{attr_name}")
            self.assertIsInstance(getattr(user.boss_elite_medals, attr_name), attr_type,
                                  msg=f"Asserting User.boss_elite_medals.{attr_name}")
            # self.assertIsInstance(
            #     getattr(user.race_medals, attr_name), attr_type,
            #     msg=f"Asserting User.race_medals.{attr_name}"
            # )

        check_instance = [
            ("first", int), ("second", int), ("third", int), ("top_10", int), ("top_20", int),
            ("top_40", int), ("top_60", int)
        ]
        for attr_name, attr_type in check_instance:
            self.assertIsInstance(getattr(user.ct_local_medals, attr_name), attr_type,
                                  msg=f"Asserting User.ct_local_medals.{attr_name}")

        check_instance = [
            ("top_25", int), ("top_100", int), ("top_1_percent", int), ("top_10_percent", int), ("top_25_percent", int),
            ("top_50_percent", int), ("top_75_percent", int)
        ]
        for attr_name, attr_type in check_instance:
            self.assertIsInstance(getattr(user.ct_global_medals, attr_name), attr_type,
                                  msg=f"Asserting User.ct_global_medals.{attr_name}")

        # Gameplay stats

        expected_results_ge = [
            (btd6.Tower.BRICKELL, 93), (btd6.Tower.ADORA, 68), (btd6.Tower.BENJAMIN, 1189),
            (btd6.Tower.ETIENNE, 108), (btd6.Tower.GERALDO, 151), (btd6.Tower.GWENDOLIN, 411),
            (btd6.Tower.OBYN, 499), (btd6.Tower.PAT_FUSTY, 143), (btd6.Tower.PSI, 204),
            (btd6.Tower.QUINCY, 125), (btd6.Tower.SAUDA, 534), (btd6.Tower.STRIKER_JONES, 59),
            (btd6.Tower.EZILI, 257), (btd6.Tower.CHURCHILL, 63),
        ]
        for hero, times_placed in expected_results_ge:
            self.assertIn(hero, user.heroes_placed.keys(),
                          msg=f"Assert {hero.value} appears in User.heroes_placed")
            self.assertGreaterEqual(user.heroes_placed[hero], times_placed,
                                    msg=f"Assert correctness of User.heroes_placed[{hero.value}]")

        self.assertIsInstance(user.stats.most_experienced_monkey, btd6.Tower,
                              msg=f"Asserting User.stats.most_experienced_monkey")
        expected_results_ge = [
            ("cash_earned", 711700673), ("challenges_completed", 198), ("collection_chests_opened", 293),
            ("coop_cash_given", 78006898), ("daily_rewards", 339), ("game_count", 9252), ("games_won", 2618),
            ("highest_round", 276), ("highest_round_chimps", 127), ("highest_round_deflation", 100),
            ("insta_monkey_collection", 796), ("monkey_teams_wins", 11), ("powers_used", 3811),
            ("total_odysseys_completed", 26), ("total_odyssey_stars", 100), ("total_trophies_earned", 3288),
            ("necro_bloons_reanimated", 476791), ("bloons_leaked", 557788),
        ]
        for attr_name, attr_expected_min in expected_results_ge:
            self.assertGreaterEqual(getattr(user.stats, attr_name), attr_expected_min,
                                    msg=f"Asserting User.stats.{attr_name}")

        expected_results_ge = [
            ("total", 455698668), ("total_coop", 26290553), ("camos", 31764834), ("regrows", 293382673),
            ("purples", 1133115), ("leads", 13291249), ("ceramics", 210121378), ("moabs", 1324838), ("bfbs", 252424),
            ("zomgs", 34974), ("bads", 3824), ("golden", 69),
        ]
        for attr_name, attr_expected_min in expected_results_ge:
            self.assertGreaterEqual(getattr(user.stats.bloons_popped, attr_name), attr_expected_min,
                                    msg=f"Asserting User.stats.bloons_popped.{attr_name}")

    def test_user_fail(self) -> None:
        """
        Test that the correct exception is raised when attempting to fetch a nonexistant user.
        """
        user_id = "blatantlywronguserid"
        correct_exception = False
        try:
            btd6.User(user_id, eager=True)
        except btd6.NotFound:
            correct_exception = True
        self.assertTrue(correct_exception, msg="Wrong user IDs should raise bloonspy.exceptions.NotFound")


if __name__ == '__main__':
    unittest.main()



import unittest
from datetime import datetime
import requests
from bloonspy import btd6


class TestBoss(unittest.TestCase):
    def test_boss(self) -> None:
        """
        Test that a boss is loaded correctly.
        """
        bosses = requests.get("https://data.ninjakiwi.com/btd6/bosses")
        boss_id = bosses.json()["body"][0]["id"]
        boss_event = btd6.BossEvent(boss_id)
        boss_standard = boss_event.standard()
        boss_elite = boss_event.elite()

        check_instance = [
            ("id", str), ("name", str), ("start", datetime), ("end", datetime), ("total_scores_standard", int),
            ("total_scores_elite", int), ("boss_banner", str), ("boss_bloon", btd6.BossBloon)
        ]
        for attr_name, attr_type in check_instance:
            self.assertIsInstance(getattr(boss_event, attr_name), attr_type,
                                  msg=f"Assert if BossEvent.{attr_name} is {attr_type}")

        check_instance = [
            ("total_scores", int), ("upvotes", int), ("plays_unique", int)
        ]
        for attr_name, attr_type in check_instance:
            self.assertIsInstance(getattr(boss_standard, attr_name), attr_type,
                                  msg=f"Assert if Boss.{attr_name} is {attr_type}")
            self.assertIsInstance(getattr(boss_elite, attr_name), attr_type,
                                  msg=f"Assert if Boss.{attr_name} is {attr_type}")

        self.assertFalse(boss_standard.is_elite)
        self.assertTrue(boss_elite.is_elite)

    def test_boss_expired(self) -> None:
        """
        Test that the correct exception is raised when attempting to fetch a nonexistant/expired boss.
        """
        boss_id = "Vortex10_l7si69ji"
        correct_exception = False
        try:
            btd6.BossEvent(boss_id, eager=True)
        except btd6.NotFound:
            correct_exception = True
        self.assertTrue(correct_exception, msg="Wrong/expired boss IDs should raise bloonspy.exceptions.NotFound")


if __name__ == '__main__':
    unittest.main()



import requests
from typing import Dict, Any, List, Union
import sys


API_URL = "https://data.ninjakiwi.com"


def get(endpoint: str, params: Dict[str, Any] = None) -> Union[List[Dict[str, Any]], Dict[str, Any]]:
    if "unittest" in sys.modules.keys():
        print(f"GET {endpoint}, {params=}")

    if params is None:
        params = {}

    resp = requests.get(API_URL + endpoint, params=params, headers={"User-Agent": "bloonspy Python Library"})
    if resp.status_code >= 500:
        if "unittest" in sys.modules.keys():
            print(resp.content)
            print(resp.headers)
        raise Exception("Server error occurred")
    if resp.status_code >= 400:
        raise Exception("Bad request")
    if "application/json" not in resp.headers.get("content-type").lower():
        raise Exception("Content is not JSON")

    data = resp.json()
    if not data["success"]:
        raise Exception(data["error"])

    return data["body"]


def get_lb_page(endpoint: str, page_num: int):
    try:
        return get(endpoint, params={"page": page_num})
    except Exception as exc:
        if str(exc) == "No Scores Available":
            return []
        raise exc



from typing import Dict, Any
from enum import Enum
from ...exceptions import NotFound
from ...utils.decorators import fetch_property
from ..Loadable import Loadable
from ..Asset import Asset
from .User import User


class TeamStatus(Enum):
    PUBLIC = "Public"
    PRIVATE = "Closed"
    INVITE_ONLY = "Invite Only"
    DISBANDED = "Disbanded"

    @staticmethod
    def from_string(value: str) -> "TeamStatus":
        status_switch = {
            "OPEN": TeamStatus.PUBLIC,
            "CLOSED": TeamStatus.PRIVATE,
            "FILTERED": TeamStatus.INVITE_ONLY,
            "DISBANDED": TeamStatus.DISBANDED,
        }
        return status_switch[value] if value in status_switch else None


class Team(Loadable):
    """A BTD6 Team."""
    endpoint = "/btd6/guild/{}"

    def _handle_exceptions(self, exception: Exception) -> None:
        error_msg = str(exception)
        if error_msg == "Invalid guild ID":
            raise NotFound(error_msg)

    def _parse_json(self, raw_resource: Dict[str, Any]) -> None:
        self._loaded = False

        copy_keys = ["name", "numMembers"]
        for key in copy_keys:
            self._data[key] = raw_resource[key]
        assets = [
            ("banner", "bannerURL"), ("icon", "iconURL"), ("frame", "frameURL"),
        ]
        for asset_name, asset_url in assets:
            self._data[asset_name] = Asset(raw_resource[asset_name], raw_resource[asset_url])
        self._data["status"] = TeamStatus.from_string(raw_resource["status"])
        self._data["owner_id"] = raw_resource["owner"].split("/")[-1]

        self._loaded = True

    @property
    @fetch_property(Loadable.load_resource)
    def name(self) -> str:
        """The name of the team."""
        return self._data["name"]

    @property
    @fetch_property(Loadable.load_resource)
    def member_count(self) -> int:
        """The team's member count."""
        return self._data["numMembers"]

    @property
    @fetch_property(Loadable.load_resource)
    def status(self) -> TeamStatus:
        """The team's entry status (public, closed, ...)."""
        return self._data["status"]

    @property
    @fetch_property(Loadable.load_resource)
    def banner(self) -> Asset:
        """The team's equipped banner."""
        return self._data["banner"]

    @property
    @fetch_property(Loadable.load_resource)
    def icon(self) -> Asset:
        """The team's equipped icon."""
        return self._data["icon"]

    @property
    @fetch_property(Loadable.load_resource)
    def frame(self) -> Asset:
        """The team's equipped frame."""
        return self._data["frame"]

    @property
    @fetch_property(Loadable.load_resource)
    def owner_id(self) -> str:
        """The ID of the user who currently owns the team."""
        return self._data["owner_id"]

    @fetch_property(Loadable.load_resource)
    def owner(self) -> User or None:
        """Fetch the owner of the team.

        .. warning::
           This function needs the property :attr:`~bloonspy.model.btd6.Team.owner_id` to be
           loaded, or it will make another API call to fetch that first.

        :return: The owner of the team.
        :rtype: User or None
        """
        if self.owner_id is None:
            return None
        return User(self.owner_id, eager=True)



import unittest
import random
from datetime import datetime
from bloonspy import Client, ChallengeFilter


class TestChallengeBrowser(unittest.TestCase):
    def test_challenge_browser(self) -> None:
        """
        Test getting the newest challenges.
        """
        challenges = Client.challenges(ChallengeFilter.NEWEST, start_from_page=2, pages=2)
        self.assertGreater(len(challenges), 0)
        some_challenge = challenges[random.randint(0, len(challenges)-1)]

        check_instance = [
            ("name", str), ("created_at", datetime), ("wins", int), ("upvotes", int)
        ]
        for attr_name, attr_type in check_instance:
            self.assertIsInstance(
                getattr(some_challenge, attr_name), attr_type,
                msg=f"Assert if Challenge.{attr_name} is {attr_type}"
            )

    def test_eager(self) -> None:
        """
        Test getting the newest challenges, fully loaded.
        """
        challenges = Client.challenges(ChallengeFilter.NEWEST, eager=True)
        self.assertGreater(len(challenges), 0)

        for challenge in challenges:
            self.assertTrue(challenge.loaded)


if __name__ == '__main__':
    unittest.main()



from typing import Dict, Any, List
from datetime import datetime
from ..utils.dictionaries import has_all_keys
from ..utils.decorators import fetch_property, exception_handler
from ..utils.api import get
from ..exceptions import NotFound


class Event:
    """A game event."""

    event_endpoint: str = "/..."
    event_dict_keys: List[str] = ["name", "start", "end"]
    event_name: str = "event"

    def __init__(self, event_id: str, eager: bool = True, event_json: Dict[str, Any] = None):
        self._id = event_id
        self._data = {}
        self._event_loaded = False

        if event_json and has_all_keys(event_json, self.event_dict_keys):
            self._parse_event(event_json)
        if eager and not self._event_loaded:
            self.load_event()

    def handle_exceptions(self, exception: Exception) -> None:
        return self._handle_exceptions(exception)

    def _handle_exceptions(self, exception: Exception) -> None:
        error_msg = str(exception)
        if error_msg == f"No f{self.event_name} with that ID exists":
            raise NotFound(error_msg)
        raise exception

    @exception_handler(handle_exceptions)
    def load_event(self, only_if_unloaded: bool = True) -> None:
        """Load the event.

        :param only_if_unloaded: Only make an API call if the event is unloaded.
            If `False`, it essentially "reloads" the event.
        :type only_if_unloaded: :class:`bool`

        :raises bloonspy.exceptions.NotFound: If the resource is not found.
        """
        if self._event_loaded and only_if_unloaded:
            return

        self._event_loaded = False

        event_list = get(self.event_endpoint)
        for event in event_list:
            if event["id"] == self._id:
                self._parse_event(event)
                return

        raise NotFound(f"No {self.event_name} with that ID exists")

    def _parse_event(self, data: Dict[str, Any]) -> None:
        self._data["name"] = data["name"]
        self._data["start"] = datetime.fromtimestamp(data["start"]/1000)
        self._data["end"] = datetime.fromtimestamp(data["end"]/1000)
        self._event_loaded = True

    @staticmethod
    def _should_load_property(key_name: str) -> callable:
        def _inner(self: "Event") -> bool:
            return key_name not in self._data
        return _inner

    @property
    def id(self) -> str:
        """The unique ID of the event."""
        return self._id

    @property
    @fetch_property(load_event, should_load=_should_load_property("name"))
    def name(self) -> str:
        """The name of the event."""
        return self._data["name"]

    @property
    @fetch_property(load_event, should_load=_should_load_property("start"))
    def start(self) -> datetime:
        """When the event starts."""
        return self._data["start"]

    @property
    @fetch_property(load_event, should_load=_should_load_property("end"))
    def end(self) -> datetime:
        """When the event ends."""
        return self._data["end"]
